/************************************************************************
					   biquandle search 
						 A. Bartholomew

                   coding started March 2008
                   
March 2014: changed code to be based on the map F rather than S, to
correspond the Generalised Biquandles for Generalised Knot Theories. 

December 2014: added capability to look for doodle switches.  These are 
switches that have order 2 and do not satisfy the S-Yang-Baxter equation.
	   

****************************************************************************/
using namespace std;

#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <list>
#include <ctype.h>
#include <csignal>
#include <stdio.h>
#include <vector>
#include <valarray>
#include <iomanip>

ofstream    debug;

#include <util.h>
#include <matrix.h>

#define OFF				0
#define BASIC			1
#define INTERMEDIATE	2
#define DETAIL 			3

#define PLAIN_FORMAT			0
#define TeX_FORMAT				1
#define BRAID_PROGRAMME_FORMAT 	2

unsigned int DEBUG = OFF;

bool BIRACK_SEARCH = false;
bool BRAID_OUTPUT_DISPLAY = false;
bool CHECK_ROWS_ARE_PERMUTATIONS = false;
bool COMPARE_INPUT_FILE = false;
bool COMPARE_S_AND_F_BIRACKS = false;
bool CONVERT_MAPS_s_TO_f = false;
bool CONVERT_MAPS_f_TO_s = false;
bool CONVERT_X_n_FORMAT = false;
bool DISPLAY_RELATED_QUANDLE = false;
bool DISPLAY_ROWS_AS_PERMUTATIONS = false;
bool DOODLE_SEARCH = false;
bool EXCLUDE_INPUT = false;
bool EXCLUDE_PERMUTATIONS = false;
bool INDECOMPOSABLE_RACKS_ONLY = false;
bool INVESTIGATING_YB = false;
bool NO_SYMMETRY_CHECK = false;
bool NUMBER_FROM_ZERO = false;
bool OMIT_INVERSE_SYMMETRY = false;
bool QUANDLE_RELATIVES_SEARCH = false;
bool QUANDLE_SEARCH = false;
bool QUANDLE_SEED_FILE = false;
bool RACK_SEARCH = false;
bool READ_CALCULATED_LIST = false;
bool S_YANG_BAXTER_SEARCH = false;
bool SIGINT_RECEIVED = false;
bool SWITCH_SEARCH = true;
bool TeX_OUTPUT_DISPLAY = false;

int n; // size of finite field used in search

bool Yang_Baxter_satisfied (matrix<int>& U, matrix<int>& D, bool rack_condition = false);
bool S_Yang_Baxter_rules (matrix<int>& U, matrix<int>& D, bool rack_condition);
bool F_Yang_Baxter_rules (matrix<int>& U, matrix<int>& D,bool rack_condition);
void display(matrix<int>& M, ostream& os, int format=0);
void display_as_perms(matrix<int>& M, ostream& os, int format=0);
void debug_display (matrix<int>& M);
bool trivial_action (matrix<int>& M);
bool preserved_diagonal (matrix<int>& M);
bool biquandle_test(matrix<int>& U, matrix<int>& D);
bool equal_rows (matrix<int>& M);
int constant_columns (matrix<int>& M);
int switch_order(matrix<int>& U, matrix<int>& D);
bool rows_are_permutations (matrix<int>& M);
bool permutation_of_XxX (matrix<int>& U, matrix<int>& D);
unsigned int action_span (matrix<int>& U);
bool F_Yang_Baxter_rule_check (matrix<int>& Us, matrix<int>& Ds);
bool S_Yang_Baxter_rule_check (matrix<int>& Uf, matrix<int>& Df);
bool power_2(matrix<int>& U, matrix<int>& D);

struct Switch 
{

	Switch(matrix<int> u,matrix<int> d):U(u),D(d){}
	matrix<int> U;
	matrix<int> D;
};

int check_list( matrix<int>& U, matrix<int>& D, list<Switch>& list_to_check, bool add=true, bool swap_trivial_D_action=true);
Switch S_sideways_map_bottom_up(matrix<int>& old_U,matrix<int> old_D);
Switch F_sideways_map_left_right(matrix<int>& old_U,matrix<int> old_D);

void sigint_handler (int sig) 
{
	cout << "\n\nWriting output file on receipt of SIGINT..." << endl;
	SIGINT_RECEIVED = true;
}

bool matrix_control::SINGLE_LINE_OUTPUT;

int main(int argc, char* argv[])
{
	bool options;
	string output_file = "biquandle.out";
	list<Switch> calculated_list;
	int discovery_count=0;
	int discovery_threshold = 1000;
	int discovery_threshold_count = 0;

	signal(SIGINT, sigint_handler); // register the SIGINT handler

	n=0; 
	
	if (argc >= 2)
	{
		if ( *(argv[1]) == '-')
			options = true;

		if (options)
		{
			char* dptr = strchr(argv[1], '#');
			if (dptr != 0)
			{
				DEBUG = BASIC;
				if (isdigit(*++dptr))
				{
					if (*dptr == '2')
						DEBUG = INTERMEDIATE;
					else if (*dptr == '3')
						DEBUG = DETAIL;
				}
			}

			if (strchr(argv[1], 'B'))
				 BRAID_OUTPUT_DISPLAY = true;


			if (strchr(argv[1], 'b'))
				BIRACK_SEARCH = true;

			if (strchr(argv[1], 'c'))
				COMPARE_INPUT_FILE = true;

			if (strchr(argv[1], 'C'))
				CONVERT_X_n_FORMAT = true;

			if (strchr(argv[1], 'd'))
			{
				DOODLE_SEARCH = true;
				S_YANG_BAXTER_SEARCH = true;
			}


			if (strchr(argv[1], 'D'))
				DISPLAY_RELATED_QUANDLE = true;
				
			if (strchr(argv[1], 'e'))
			{
				COMPARE_INPUT_FILE = true;
				EXCLUDE_INPUT = true;
			}

			if (strchr(argv[1], 'f'))
			{
				QUANDLE_SEED_FILE = true;
				READ_CALCULATED_LIST = true;
			}

			if (strchr(argv[1], 'F'))
			{
				CONVERT_MAPS_s_TO_f = true;
				READ_CALCULATED_LIST = true;
			}


			if (strchr(argv[1], 'g'))
			{
				CONVERT_MAPS_f_TO_s = true;
				READ_CALCULATED_LIST = true;
			}

			if (strchr(argv[1], 'i'))
			{
				/* added to allow the filtering of indecomposable quandles from a calculated_list
				   read from an input file.  The name has been chosen to provide for an extension 
				   of the concept to biracks in the future, if necessary.
				*/
				INDECOMPOSABLE_RACKS_ONLY = true;
			}

			if (strchr(argv[1], 'I'))
			{
				OMIT_INVERSE_SYMMETRY = true;
			}

			if (strchr(argv[1], 'M'))
			{
				COMPARE_INPUT_FILE = true;
				COMPARE_S_AND_F_BIRACKS = true;
			}

			if (strchr(argv[1], 'p'))
				CHECK_ROWS_ARE_PERMUTATIONS = true;

			if (strchr(argv[1], 'P'))
				DISPLAY_ROWS_AS_PERMUTATIONS = true;

			if (strchr(argv[1], 'q'))
			{
				QUANDLE_SEARCH = true;
				BIRACK_SEARCH = true;
				READ_CALCULATED_LIST = true;
			}

			if (strchr(argv[1], 'Q'))
			{
				QUANDLE_RELATIVES_SEARCH = true;
			}

			if (strchr(argv[1], 'r'))
			{
				RACK_SEARCH = true;
				BIRACK_SEARCH = true;
				READ_CALCULATED_LIST = true;
			}

			if (strchr(argv[1], 'R'))
				READ_CALCULATED_LIST = true;

			if (strchr(argv[1], 's'))
				SWITCH_SEARCH = false;

			if (strchr(argv[1], 'S'))
			{
				/* The S-Yang-Baxter provides backwards compatibility with earlier versions of the code that followed
				   the birack theory developed from the switch operation S.  By setting S_YANG_BAXTER_SEARCH true
				   the YB rules resulting from S(x,y) = (y^x,x_y) are used rather than those from the current theory
				   developed from F(x,y) = (y_x,x^y).  It does this by changing the function Yang_Baxter_satisfied
				   calls from F_Yang_Baxter_rules to S_Yang_Baxter_rules.
				   
				   Note that rack, quandle and biquandle searches also check YB rules but by comparing Reidemeister III moves
				   using F and S operations it can be seen that the rack condition is the same in both cases 
				   (as one would expect given that the sideways approach only affects the down operation, which is trivial 
				   in the case of a rack).  The biquandle_test also uses the appropriate biquandle condition based on 
				   S_YANG_BAXTER_SEARCH.
				*/
				S_YANG_BAXTER_SEARCH = true;
			}

			if (strchr(argv[1], 'T'))
				TeX_OUTPUT_DISPLAY = true;

			if (strchr(argv[1], 'x'))
				EXCLUDE_PERMUTATIONS = true;

			if (strchr(argv[1], 'X'))
				NO_SYMMETRY_CHECK = true;

			if (strchr(argv[1], 'n'))
			{
	    		char* c1 = strchr(argv[1],'n')+1;
	    		if (*c1 == '=')
	    		{
					char* c2 = ++c1 + 1;
					while(isdigit(*c2))
			    		c2++;
					get_number(n,c1);
	    		}
			}

			if (strchr(argv[1], 'Y'))
			{
				INVESTIGATING_YB = true;
			}

			if (strchr(argv[1], 'Z'))
				NUMBER_FROM_ZERO = true;
		}

	}
	else if (argc == 1 || argc > 3)
	{
		cout << "\n\nUsage: biquandle-search -n=<n>[#[1|2|3]BbCcdDefFIipPQqRrSsTXxZ [<input_file_1>] [<input_file_2>]]" << endl;
		cout << "      n=<n> specify the size of the field (must be provided)" << endl;		
		cout << "      #[1|2|3] generate debug output" << endl;
		cout << "        optionally specify level: 1=BASIC, 2=INTERMEDIATE, 3=DETAIL (default BASIC)" << endl;
		cout << "      B display matrix output in braid programme switch format" << endl;
		cout << "      b birack (not biquandle) search" << endl;
		cout << "      C convert the specified Z option on input the opposite on output" << endl;
		cout << "      c compare search results with <input_file_2> (note <input_file_1> need not be present)" << endl;
		cout << "      d doodle switch search (may be combined with b to avoid testing the biquandle condition)" << endl;
		cout << "      D display related quandle information as part of TeX output (requires quandle list in <input_file_2>)" << endl;
		cout << "      e exclude anything isomorphic to an item in <input_file_2> (turns on c, <input_file_1> need not be present)" << endl;
		cout << "      f use a quandle seed file for the search (turns on R)" << endl;
		cout << "      F produce sideways maps from the S-biracks in <input_file_1> (turns on R)" << endl;
		cout << "      I do not remove inverses in symmetry check" << endl;
		cout << "      i identify indecomposable racks only" << endl;
		cout << "      P display rows as permutations" << endl;
		cout << "      p check rows of the list read from <input_file_1> are permutations" << endl;
		cout << "      Q search for quandle relatives in calculated list (may be used with R)" << endl;
		cout << "      q search for quandles (turns on b and R)" << endl;
		cout << "      R read data from <input_file_1>" << endl;
		cout << "      r search for racks (turns on b and R)" << endl;
		cout << "      S S-Yang-Baxter search" << endl;
		cout << "      s start search from birack conditions not switch conditions" << endl;
		cout << "      T display matrix output in TeX format" << endl;
		cout << "      X exclude summetry check" << endl;
		cout << "      x exclude permutations from the isomorphic symmetry check" << endl;
		cout << "      Y only run code investigating Y-B equation" << endl;
		cout << "      Z input and output X_n={0,...,n-1} rather than X_n={1,...,n}" << endl;
		exit(0);
	}
	
	if (!n)
	{
		cout << "\nError!  No value provided for n." << endl;
		exit(0);
	}

if (DEBUG >= BASIC)
{
    /* establish a debug file */
    debug.open ("biquandle.dbg");

    if (!debug)
    {
        cout << "\nError opening debug file";
        exit(0);
    }

	debug << "\nn = " << n << endl;
	debug << "BIRACK_SEARCH = " << (BIRACK_SEARCH? "true": "false") << endl;
	debug << "CHECK_ROWS_ARE_PERMUTATIONS = " << (CHECK_ROWS_ARE_PERMUTATIONS? "true": "false") << endl;
	debug << "COMPARE_INPUT_FILE = " << (COMPARE_INPUT_FILE? "true": "false") << endl;
	debug << "COMPARE_S_AND_F_BIRACKS = " << (COMPARE_S_AND_F_BIRACKS? "true": "false") << endl;
	debug << "CONVERT_MAPS_s_TO_f = " << (CONVERT_MAPS_s_TO_f? "true": "false") << endl;
	debug << "CONVERT_MAPS_f_TO_s = " << (CONVERT_MAPS_f_TO_s? "true": "false") << endl;
	debug << "CONVERT_X_n_FORMAT = " << (CONVERT_X_n_FORMAT? "true": "false") << endl;
	debug << "DISPLAY_ROWS_AS_PERMUTATIONS = " << (DISPLAY_ROWS_AS_PERMUTATIONS? "true": "false") << endl;
	debug << "EXCLUDE_INPUT = " << (EXCLUDE_INPUT? "true": "false") << endl;
	debug << "EXCLUDE_PERMUTATIONS = " << (EXCLUDE_PERMUTATIONS? "true": "false") << endl;
	debug << "INDECOMPOSABLE_RACKS_ONLY = " << (INDECOMPOSABLE_RACKS_ONLY? "true": "false") << endl;
	debug << "INVESTIGATING_YB = " << (INVESTIGATING_YB? "true": "false") << endl;
	debug << "NO_SYMMETRY_CHECK = " << (NO_SYMMETRY_CHECK? "true": "false") << endl;
	debug << "NUMBER_FROM_ZERO = " << (NUMBER_FROM_ZERO? "true": "false") << endl;
	debug << "OMIT_INVERSE_SYMMETRY = " << (OMIT_INVERSE_SYMMETRY? "true": "false") << endl;
	debug << "QUANDLE_RELATIVES_SEARCH = " << (QUANDLE_RELATIVES_SEARCH? "true": "false") << endl;
	debug << "QUANDLE_SEARCH = " << (QUANDLE_SEARCH? "true": "false") << endl;
	debug << "QUANDLE_SEED_FILE = " << (QUANDLE_SEED_FILE? "true": "false") << endl;
	debug << "RACK_SEARCH = " << (RACK_SEARCH? "true": "false") << endl;
	debug << "READ_CALCULATED_LIST = " << (READ_CALCULATED_LIST? "true": "false") << endl;
	debug << "S_YANG_BAXTER_SEARCH = " << (S_YANG_BAXTER_SEARCH? "true": "false") << endl;
	debug << "SWITCH_SEARCH = " << (SWITCH_SEARCH? "true": "false") << endl;
	debug << "TeX_OUTPUT_DISPLAY = " << (TeX_OUTPUT_DISPLAY? "true": "false") << endl;
}

	/* prepare output file */
	ofstream output(output_file.c_str());

	if (!output)
	{
		cout << "\nError opening output file";
		exit(0);
	}

	/* The (x,y) entry of U and D are y^x and y_x, respectively so that 
	   the birack conditions require rows of U and D to be permutations.
	*/
	matrix<int> U(n,n), D(n,n);
	
	/* the matrix Identity is only used when considering quandle-related biracks */
	matrix<int> Identity(n,n);
	for (int i=0; i< n; i++)
	for (int j=0; j< n; j++)
		Identity[i][j] = j;
	
	if (READ_CALCULATED_LIST)
	{
		/* read in the list of switches/biracks/biquandles from the <input_file_1>, 
		   the only requirement being that the matrices are specified as U =...
		   so we can look for a U and skip any B^2_{1} etc.
		*/
		ifstream input(argv[2]);
		if (input)
		{
			string next_line;
			while (getline(input,next_line))
			{
				istringstream iss(next_line);
				char ch;
				do
				{
					iss >> ch;	
				} while (!(ch == 'U'));
								
				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					do
					{
						iss >> ch;	
						if (isdigit(ch))
							iss.putback(ch);
					} while (!isdigit(ch));
					iss >> U[i][j];
					
					/* The code is written with X_n={0,...,n-1} so if the
					   input is not in that form adjust it */
					if (!NUMBER_FROM_ZERO)
						U[i][j]--;						
				}

				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					do
					{
						iss >> ch;	
						if (isdigit(ch))
							iss.putback(ch);
					} while (!isdigit(ch));
					iss >> D[i][j];

					if (!NUMBER_FROM_ZERO)
						D[i][j]--;
				}

if (DEBUG >= BASIC)
{
	debug << "\nRead from input file: U = ";
	debug_display (U);
	debug << " D = ";
	debug_display (D);
	debug << endl;
}
				if (CONVERT_MAPS_s_TO_f)
				{
					/* The input list is assumed to be a list of S-biracks and we convert each one to the corresponding
					   sideways map so we produce a list of sideways maps that is aligned from a naming perspective with 
					   the input list.
					*/				
					calculated_list.push_back(S_sideways_map_bottom_up(U,D));		
				}
				else if (CONVERT_MAPS_f_TO_s)
				{
					/* The input list is assumed to be a list of F-biracks and we convert each one to the corresponding
					   switch so we produce a list of switches that is aligned from a naming perspective with 
					   the input list.
					*/				
					calculated_list.push_back(F_sideways_map_left_right(U,D));		
				}
				else if (QUANDLE_SEED_FILE)
				{

if (DEBUG >= BASIC)
	debug << "Assuming U is a quandle: enumarating rows of D as permutations" << endl;
	
					/* Write U to stdout as an indication or progress */
					display(U, cout);
					cout << endl;
	
					/* We are only interested in the up matrix we have read, which is assumed to be a quandle.
					   We carry out the birack-based search considering only different possibilities for the down
					   matrix.

					   Start from the premise that we have a birack,  so that every row 
					   of U and D is a permutation
					*/
					for (int i=0; i< n; i++)
					for (int j=0; j< n; j++)
						D[i][j] = j;
						
					vector<int> Dv(n*n);	
					
					for (int i=0; i< n*n; i++)
						Dv[i] = i%n;
					
					bool D_enumeration_complete = false;
					do
					{
				
if (DEBUG >= BASIC)
{
	debug << " D = ";
	debug_display (D);
	debug << endl;
}

if (DEBUG >= DETAIL)
{
	debug << " Dv = ";
	for (int i=0; i< n*n; i++)
		debug << Dv[i] << ' ';
	debug << endl;
}

						if ( permutation_of_XxX(U,D)
						     && (BIRACK_SEARCH || biquandle_test(U,D)) 
						     && (DOODLE_SEARCH? power_2(U,D) && !Yang_Baxter_satisfied(U,D): Yang_Baxter_satisfied(U,D))
						   )
						{
							cout << "!" << flush;
							if (discovery_count++ > discovery_threshold)
							{
								discovery_threshold_count++;
								discovery_count = 0;
								cout << discovery_threshold*discovery_threshold_count << flush;
							}

if (DEBUG >= BASIC)
{
	if (DOODLE_SEARCH)
		debug << "  found doodle switch" << endl;
	else
		debug << "  found " << (BIRACK_SEARCH? "birack": "biquandle") << endl;	
}
				
							if (NO_SYMMETRY_CHECK)
								calculated_list.push_back(Switch(U,D));		
							else				
								check_list(U,D,calculated_list);
			
						}
				
						/* increment D */			
						for (int i=n; i>0; i--)
						{
							if (next_permutation(Dv.begin()+(i-1)*n, Dv.begin()+i*n))
							{
								for (int j=0; j< n; j++)
									D[i-1][j] = Dv[(i-1)*n+j];
								break;
							}
							else
							{					
								for (int j=0; j< n; j++)
									Dv[(i-1)*n+j] = D[i-1][j] = j;
			
								if (i==1)
									D_enumeration_complete = true;
							}
						}
								
					} while (!D_enumeration_complete && !SIGINT_RECEIVED);
				}
				else if ((BIRACK_SEARCH || biquandle_test(U,D)) && (DOODLE_SEARCH? power_2(U,D) && !Yang_Baxter_satisfied(U,D): Yang_Baxter_satisfied(U,D)))
				{
					if (CHECK_ROWS_ARE_PERMUTATIONS)
					{
						/* Check that the rows of the up matrix are permuations */
						if (!rows_are_permutations (U))
						{
				
if (DEBUG >= BASIC)
{
	debug << "Counterexample found for switch map up operation not being a permutation" << endl;	
}
						cout << "Counterexample found for switch map up operation not being a permutation" << endl;
						exit(0);
						}
					}

					cout << "!" << flush;
					if (discovery_count++ > discovery_threshold)
					{
						discovery_threshold_count++;
						discovery_count = 0;
						cout << discovery_threshold*discovery_threshold_count << flush;
					}

if (DEBUG >= BASIC)
{
	if (DOODLE_SEARCH)
		debug << "  found doodle switch" << endl;
	else
		debug << "  found " << (BIRACK_SEARCH? "birack": "biquandle") << endl;	
}

					if (NO_SYMMETRY_CHECK)
						calculated_list.push_back(Switch(U,D));		
					else				
						check_list(U,D,calculated_list);
				}
			}
		}
		else
		{
	        cout << "\nError opening " << argv[2] << endl;
    	    exit(0);
		}
	}
	else if (SWITCH_SEARCH)
	{
		/* start with permutations of X x X.  Consider a matrix of pairs where the (x,y) entry is given by F(x,y),
		   or S(x,y) if S_YANG_BAXTER_SEARCH is true.  We populate this matrix with all permutations of XxX and for 
		   each one derive the corresponding U and D.
		   Since F(a,b) = (b_a,a^b) (or S(a,b) = (b^a, a_b) ) and we record U and D with the (x,y) entry as y^x amd y_x 
		   respectively, we create D from the transpose of the "down" components of the permuation.
		*/
		vector<int> perm(n*n);
		int k=0;	
		for (int i=0; i< n; i++)
		for (int j=0; j< n; j++)
			perm[k++]=j*n+i;
			
		do
		{

if (DEBUG >= BASIC)
{
	debug << "perm: ";
	for (int i=0; i< n*n; i++)
		debug << perm[i] << " ";
	debug << endl;	
}
			for (int i=0; i< n; i++)
			for (int j=0; j< n; j++)
				U[i][j] = perm[n*i+j]/n;
			
			for (int i=0; i< n; i++)
			for (int j=0; j< n; j++)
				D[j][i] = perm[n*i+j]%n; 

if (DEBUG >= BASIC)
{
	debug << "\nU = ";
	debug_display (U);
	debug << "D = ";
	debug_display (D);
	debug << endl;
}

			/* For U,D to be a birack the rows of U and D must be a permutation */
if (DEBUG >= BASIC)
	debug << "  birack row permutation condition for U: " << endl;
	
			for (int i=0; i<n; i++)
			{
				valarray<int> check_perm(0,n);
				for (int j=0; j< n; j++)
				{
					if (check_perm[U[i][j]])
					{
if (DEBUG >= BASIC)
{
	debug << "    row " << i << " of U is not a permutation" << endl;
	debug << "    fails" << endl;
}					


						/* We have found that the current value of perm[n*i+j] results in 
						   row i of U not being a perm, so rearrange perm so that the next 
						   permutation increases the value of perm[n*i+j] to it's next 
						   lexicographic value.  This means that we have to arrange all 
						   the values from perm[n*i+j+1] to the end in decreasing order.
						*/
						sort((vector<int>::iterator)&perm[n*i+j+1],perm.end(),greater<int>());

if (DEBUG >= BASIC)
{
	debug << "reset perm to: ";
	for (int i=0; i< n*n; i++)
		debug << perm[i] << " ";
	debug << endl;	
}						
						goto increment_switch_search; // not a permutation
					}
					else
						check_perm[U[i][j]] = 1;
				}				
			}

if (DEBUG >= BASIC)
{
	debug << "    OK" << endl; // still here, not jumped to next permutation
	debug << "  birack row permutation condition for D: " << endl;
}

			for (int i=0; i<n; i++)
			{
				valarray<int> check_perm(0,n);
				for (int j=0; j< n; j++)
				{
					if (check_perm[D[i][j]])
					{
if (DEBUG >= BASIC)
{
	debug << "    row " << i << " of D is not a permutation\n" << endl;
	debug << "    fails" << endl;
}					

						/* Note that the transpose of D is set from perm.  We have found 
						   that the current value of perm[n*j+i] results in row i of D 
						   (column i of D^T) not being a perm, so rearrange perm so that the next 
						   permutation increases the value of perm[n*j+i] to it's next 
						   lexicographic value.  This means that we have to arrange all 
						   the values from perm[n*j+i+1] to the end in decreasing order.
						*/
						sort((vector<int>::iterator)&perm[n*j+i+1],perm.end(),greater<int>());


if (DEBUG >= BASIC)
{
	debug << "reset perm to: ";
	for (int i=0; i< n*n; i++)
		debug << perm[i] << " ";
	debug << endl;	
}						
						goto increment_switch_search; // not a permutation
					}
					else
						check_perm[D[i][j]] = 1;
				}				
			}

if (DEBUG >= BASIC)
	debug << "    OK" << endl;

			if ((BIRACK_SEARCH || biquandle_test(U,D)) && (DOODLE_SEARCH? power_2(U,D) && !Yang_Baxter_satisfied(U,D): Yang_Baxter_satisfied(U,D)))
			{
				cout << "!" << flush;
				if (discovery_count++ > discovery_threshold)
				{
					discovery_threshold_count++;
					discovery_count = 0;
					cout << discovery_threshold*discovery_threshold_count << flush;
				}

if (DEBUG >= BASIC)
{
	if (DOODLE_SEARCH)
		debug << "  found doodle switch" << endl;
	else
		debug << "  found " << (BIRACK_SEARCH? "birack": "biquandle") << endl;	
}

				if (NO_SYMMETRY_CHECK)
					calculated_list.push_back(Switch(U,D));		
				else				
					check_list(U,D,calculated_list);
	
			}
	
		increment_switch_search:;
	
		} while (!SIGINT_RECEIVED && next_permutation(perm.begin(),perm.end()));		
	}
	else
	{
		/* start from the premise that we have a birack,  so that every row 
		   of U and D is a permutation
		*/
		for (int i=0; i< n; i++)
		for (int j=0; j< n; j++)
			U[i][j] = D[i][j] = j;
			
		vector<int> Uv(n*n);
		vector<int> Dv(n*n);	
		
		for (int i=0; i< n*n; i++)
			Uv[i] = Dv[i] = i%n;
	
		bool D_enumeration_complete = false;
		bool U_enumeration_complete = false;
		
		do
		{
	
if (DEBUG >= BASIC)
{
	debug << "\nU = ";
	debug_display (U);
	debug << " D = ";
	debug_display (D);
	debug << endl;
}

if (DEBUG >= DETAIL)
{
	debug << "\nUv = ";
	for (int i=0; i< n*n; i++)
		debug << Uv[i] << ' ';
	debug << endl;
	debug << "\nDv = ";
	for (int i=0; i< n*n; i++)
		debug << Dv[i] << ' ';
	debug << endl;
}

			if (     permutation_of_XxX(U,D)
			     && (BIRACK_SEARCH || biquandle_test(U,D)) 
			     && (DOODLE_SEARCH? power_2(U,D) && !Yang_Baxter_satisfied(U,D): Yang_Baxter_satisfied(U,D))
			   )
			{
				cout << "!" << flush;
				if (discovery_count++ > discovery_threshold)
				{
					discovery_threshold_count++;
					discovery_count = 0;
					cout << discovery_threshold*discovery_threshold_count << flush;
				}

if (DEBUG >= BASIC)
{
	if (DOODLE_SEARCH)
		debug << "  found doodle switch" << endl;
	else
		debug << "  found " << (BIRACK_SEARCH? "birack": "biquandle") << endl;	
}
				
				if (NO_SYMMETRY_CHECK)
					calculated_list.push_back(Switch(U,D));		
				else				
					check_list(U,D,calculated_list);

			}
	
			/* increment D */
			for (int i=n; i>0; i--)
			{
				if (next_permutation(Dv.begin()+(i-1)*n, Dv.begin()+i*n))
				{
					for (int j=0; j< n; j++)
						D[i-1][j] = Dv[(i-1)*n+j];
					break;
				}
				else
				{
					for (int j=0; j< n; j++)
						Dv[(i-1)*n+j] = D[i-1][j] = j;
			
					if (i==1)
						D_enumeration_complete = true;
				}
			}
			
			if (D_enumeration_complete)
			{
				/* increment U */		
				for (int i=n; i>0; i--)
				{
					if (next_permutation(Uv.begin()+(i-1)*n, Uv.begin()+i*n))
					{
						/* put out a comfort character to show programme progress */
						cout << "." << flush;
	
						for (int j=0; j< n; j++)
							U[i-1][j] = Uv[(i-1)*n+j];
						break;
					}
					else
					{
						for (int j=0; j< n; j++)
							Uv[(i-1)*n+j] = U[i-1][j] = j;
	
						if (i==1)
							U_enumeration_complete = true;
					}
				}
				D_enumeration_complete = false;
			}
					
		} while (!U_enumeration_complete && !SIGINT_RECEIVED);
			
	}

	cout << "\n\nFound " << calculated_list.size();
	
	if (DOODLE_SEARCH)
		cout << " doodle switches\n" << endl;
	else
		cout << (BIRACK_SEARCH?" biracks\n":" biquandles\n") << endl;
		
	list<Switch>::iterator lptr=calculated_list.begin();

	if (RACK_SEARCH && !CONVERT_MAPS_s_TO_f && !CONVERT_MAPS_f_TO_s)
	{
		/* a birack is a rack if it has either a trivial up or down action, so we remove any members of the 
		   calculated list that do not have a trivial action, leaving racks
		*/
				
		list<Switch>::iterator lptr=calculated_list.begin();
			
		while (lptr != calculated_list.end())
		{
			if (!trivial_action(lptr->U) && !trivial_action(lptr->D)) // if there's not a trivial action it's not a rack
				lptr = calculated_list.erase(lptr); // remove lptr from the list,
			else
				lptr++;
		}
		cout << "\nReduced list to " << calculated_list.size() << " racks" << endl;
				
	}
	else if (QUANDLE_SEARCH && !CONVERT_MAPS_s_TO_f && !CONVERT_MAPS_f_TO_s)
	{
		if (INDECOMPOSABLE_RACKS_ONLY)
		{
			/* the calculated list contains quandles and we wish to isolate those that are indecomposable.  These
			   are the quandles whose inner group acts transitively on the the set X=1,...,n.  The inner group is 
			   the group generated by the permutations given by the rows of the quandle operation (U).  This group acts
			   transitively if for each a,b in X there is a word w in the group such that a^w=b.  To test this we 
			   construct a graph with vertices 1,...,n and join vertex 1 to vertex i if there is a generator of the inner
			   group that takes 1 to i.  We then recursively extend the graph by considering all the vertices i added to the 
			   graph and see where the inner group generators take those vertices.  If when the process finishes we have reached 
			   every vertex, the action is transitive, if we have not, then it isn't.
			*/
if (DEBUG >= BASIC)
	debug << "\nIsolating indecomposable quandles in calculated list" << endl;
	
			list<Switch>::iterator lptr=calculated_list.begin();
				
			while (lptr != calculated_list.end())
			{	
//						lptr = calculated_list.erase(lptr); // this is a decomposable quandle,
				bool decomposable = (lptr->U.numcols() != action_span(lptr->U));

if (DEBUG >= BASIC)
{	
	debug << "  U = ";
	debug_display (lptr->U);
//	debug << " D = ";
//	debug_display (lptr->D);
//	debug << endl;
}

				if (decomposable)
				{
if (DEBUG >= BASIC)
	debug << " is decomposable" << endl;
	
					lptr = calculated_list.erase(lptr);
				}
				else
				{
if (DEBUG >= BASIC)
	debug << " is indecomposable" << endl;

					lptr++;			
				}

			}
			cout << "\nReduced list to " << calculated_list.size() << " quandles" << endl;
		}
		else
		{
			/* the calculated list contains biracks from which we wish to isolate the quandles. A birack 
			   is a quandle if it has either a trivial up or down action, preserves the diagonal in the 
			   corresponding down or up action and satisfies the quandle version of the Yang Baxter equations.
			*/

if (DEBUG >= BASIC)
	debug << "Looking for quandles in calculated list" << endl;
	
			list<Switch>::iterator lptr=calculated_list.begin();
				
			while (lptr != calculated_list.end())
			{	
if (DEBUG >= BASIC)
{	
	debug << "\nU = ";
	debug_display (lptr->U);
	debug << " D = ";
	debug_display (lptr->D);
	debug << endl;
}
			
				if ((!trivial_action(lptr->U) || !preserved_diagonal(lptr->D) || !Yang_Baxter_satisfied(lptr->D,Identity,true)) && 
				    (!trivial_action(lptr->D) || !preserved_diagonal(lptr->U) || !Yang_Baxter_satisfied(lptr->U,Identity,true))
				   )
					lptr = calculated_list.erase(lptr); // not a quandle,
				else
					lptr++;			
			}
			cout << "\nReduced list to " << calculated_list.size() << " quandles" << endl;
		}
	}
	else if (QUANDLE_RELATIVES_SEARCH)
	{
		/* Remove from the calculated list entries that are not related to a quandle.  The list
		   will contain biracks of some kind (possibly biquandles or racks) so we know the rows 
		   of U and D are permutations.  A birack is related to a quandle if substituting either 
		   U or D with the identity defines a quandle (or the Flip of a quandle).
		 
		   We have to check whether either U or D, together with the identity action, satsfies the 
		   Yang Baxter relation for a rack and preserves the diagonal.  If neither U not D
		   satisfies this condition we remove it from the list as it's not quandle-related. 
		*/
		
		list<Switch>::iterator lptr=calculated_list.begin();
			
		while (lptr != calculated_list.end())
		{	
			if ((!preserved_diagonal(lptr->U) || !Yang_Baxter_satisfied(lptr->U,Identity,true)) && 
			    (!preserved_diagonal(lptr->D) || !Yang_Baxter_satisfied(lptr->D,Identity,true))
			   )
					lptr = calculated_list.erase(lptr); // not related to a quandle,
			else
				lptr++;			
		}
		cout << "\nReduced list to " << calculated_list.size() << " items that are related to a quandle" << endl;
				
	}

	/* The compare list needs a wide enough scope to allow it to be used to display 
	   related quandle information later
	*/
	list<Switch> compare_list;			
	if (COMPARE_INPUT_FILE || DISPLAY_RELATED_QUANDLE)
	{
		/* we might have read data from one input file and want to compare it with another */
		ifstream input((READ_CALCULATED_LIST?argv[3]:argv[2])); 
		if (input)
		{
			matrix<int> locU(n,n), locD(n,n);
			string next_line;
			int num_extra_list_items = 0;

			/* create another list from the input file */
			discovery_count = discovery_threshold_count = 0;				
			
			while (getline(input,next_line))
			{
				istringstream iss(next_line);
				char ch;
				do
				{
					iss >> ch;	
				} while (!(ch == 'U'));
								
				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					do
					{
						iss >> ch;	
						if (isdigit(ch))
							iss.putback(ch);
					} while (!isdigit(ch));
					iss >> locU[i][j];

					if (!NUMBER_FROM_ZERO)
						locU[i][j]--;
				}

				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					do
					{
						iss >> ch;	
						if (isdigit(ch))
							iss.putback(ch);
					} while (!isdigit(ch));
					iss >> locD[i][j];

					if (!NUMBER_FROM_ZERO)
						locD[i][j]--;
				}

				cout << "." << flush;
				if (discovery_count++ > discovery_threshold)
				{
					discovery_threshold_count++;
					discovery_count = 0;
					cout << discovery_threshold*discovery_threshold_count << flush;
				}
				
				compare_list.push_back(Switch(locU,locD));
			}
			cout << "\nRead " << compare_list.size() << " items from " << (READ_CALCULATED_LIST?argv[3]:argv[2]) << " into the compare list" << endl;
			
			if (EXCLUDE_INPUT)
			{
				/* remove any members of the calculated list that appear in the compare list */
				
				list<Switch>::iterator lptr=calculated_list.begin();
				discovery_count = discovery_threshold_count = 0;				
				cout << endl;
			
				while (lptr != calculated_list.end())
				{
					cout << "*" << flush;
					if (discovery_count++ > discovery_threshold)
					{
						discovery_threshold_count++;
						discovery_count = 0;
						cout << discovery_threshold*discovery_threshold_count << flush;
					}
				
					if (check_list(lptr->U,lptr->D,compare_list,false,false)) //don't add new items to list, or swap trivial D actions.
						lptr = calculated_list.erase(lptr); // remove lptr from the calculated list, as it's in the compare list
					else
						lptr++;
				}
				cout << "\nReduced list to " << calculated_list.size() << (BIRACK_SEARCH?" biracks\n":" biquandles\n") << endl;
				
			}
			else if (COMPARE_INPUT_FILE)
			{
				/* Compare the compare list against the calculated list.
				
				   If COMPARE_S_AND_F_BIRACKS is true the calculated list is assumed to contain S-biracks
				   and the compare list F-biracks.  In this case S-biracks are converted to F-biracks using
				   S_sideways_map_bottom_up and F-biracks are converted to S-biracks using F_sideways_map_left_right
				   before the comparison is made.
				*/
if (DEBUG >= BASIC)
	debug << "\nLook for switches U', D' from the compare_list in the calculated_list" << endl;
				list<Switch>::iterator lptr=compare_list.begin();
				discovery_count = discovery_threshold_count = 0;				
				cout << endl;
				
				while (lptr != compare_list.end())
				{
					
if (DEBUG >= BASIC)
{
	debug << "\nU' = ";
	debug_display (lptr->U);
	debug << " D' = ";
	debug_display (lptr->D);
	debug << endl;
}
					cout << ">" << flush;
					if (discovery_count++ > discovery_threshold)
					{
						discovery_threshold_count++;
						discovery_count = 0;
						cout << discovery_threshold*discovery_threshold_count << flush;
					}
					
					int position = 0;
					if (COMPARE_S_AND_F_BIRACKS)
					{
						Switch S = F_sideways_map_left_right(lptr->U,lptr->D);
if (DEBUG >= BASIC)
{
	debug << "\nReplacing U' = Uf and D' = Df with the corresponding sideways map" << endl;
	debug << "Us = ";
	debug_display (S.U);
	debug << " Ds' = ";
	debug_display (S.D);
	debug << endl;
}

						position = check_list(S.U,S.D,calculated_list,false,false); //don't add new items to list, or swap trivial D actions.
					}
					else
					{
						position = check_list(lptr->U,lptr->D,calculated_list,false,false); //don't add new items to list, or swap trivial D actions.
					}
					
					if (!position)
					{
if (DEBUG >=BASIC)
	debug << "\nNot found" << endl;
						num_extra_list_items++;
					}
					else
					{
if (DEBUG >=BASIC)
	debug << "\nFound at position (numbered from 1) " << position << endl;
					}
					lptr++;
				}

				cout << "\nCompare list contains " << num_extra_list_items << (BIRACK_SEARCH?" biracks":" biquandles") <<" not in the calculated list\n" << endl;

				/* Check the calculated list against the compare list */
if (DEBUG >= BASIC)
	debug << "\nLook for switches U, D from the calculated_list in the compare_list" << endl;
				lptr=calculated_list.begin();
				discovery_count = discovery_threshold_count = 0;				
				cout << endl;
				
				num_extra_list_items = 0;
			
				while (lptr != calculated_list.end())
				{
					
if (DEBUG >= BASIC)
{
	debug << "\nU = ";
	debug_display (lptr->U);
	debug << " D = ";
	debug_display (lptr->D);
	debug << endl;
}
					cout << "<" << flush;
					if (discovery_count++ > discovery_threshold)
					{
						discovery_threshold_count++;
						discovery_count = 0;
						cout << discovery_threshold*discovery_threshold_count << flush;
					}
					
					int position = 0;
					if (COMPARE_S_AND_F_BIRACKS)
					{
						Switch F = S_sideways_map_bottom_up(lptr->U,lptr->D);
if (DEBUG >= BASIC)
{
	debug << "\nReplacing U = Us and D = Ds with the corresponding sideways map" << endl;
	debug << "Uf = ";
	debug_display (F.U);
	debug << " Df' = ";
	debug_display (F.D);
	debug << endl;
}

						position = check_list(F.U,F.D,compare_list,false,false); //don't add new items to list, or swap trivial D actions.
					}
					else
					{
						position = check_list(lptr->U,lptr->D,compare_list,false,false); //don't add new items to list, or swap trivial D actions.
					}
					
					if (!position)
					{
if (DEBUG >=BASIC)
	debug << "\nNot found" << endl;
						num_extra_list_items++;
					}
					else
					{
if (DEBUG >=BASIC)
	debug << "\nFound at position (numbered from 1) " << position << endl;
					}				
					lptr++;

				}

				cout << "\nCalculated list contains " << num_extra_list_items << (BIRACK_SEARCH?" biracks":" biquandles") << " not in the compare list\n" << endl;
			}
		}
		else
			cout << "unable to open input file" << endl;
	}

	if (CONVERT_X_n_FORMAT)
	{
		NUMBER_FROM_ZERO = !NUMBER_FROM_ZERO;

if (DEBUG >= BASIC)
	debug << "Converting X_n format, NUMBER_FROM_ZERO set to " << (NUMBER_FROM_ZERO? "true" : "false") << endl;
	}
	
	if (INVESTIGATING_YB)
	{
		/* We test the relationship between the maps F and S. If S_YANG_BAXTER_SEARCH is true
		   we have read a list of S-biracks and wish to test the F-Yang-Baxter relations.  
		   In this case we take each S-birack in turn and evaluate the S_sideways_map_bottom_up,
		   which gives us U_f and D_f.  We then check FYB(a',b',c') using U_f and D_f.

		   Otherwise we have read a list of F-biracks and wish to test the S-Yang-Baxter relations.  
		   In this case we take each F-birack in turn and evaluate the F_sideways_map_left_right,
		   which gives us U_s and D_s.  We then check SYB(a',b',c') using U_s and D_s.
		*/

if (DEBUG >= BASIC)
	debug << "\nInvestigating YB" << endl;

			
		list<Switch>::iterator lptr=calculated_list.begin();
		while (lptr != calculated_list.end())
		{
			if (S_YANG_BAXTER_SEARCH)
			{
				if (!F_Yang_Baxter_rule_check(lptr->U,lptr->D))
				{
				
if (DEBUG >= BASIC)
{
	debug << "Counterexample found for FYB(a',b',c')" << endl;	
}
					cout << "Counterexample found for FYB(a',b',c')" << endl;
					exit(0);
				}
			}
			else
			{
				if (!S_Yang_Baxter_rule_check(lptr->U,lptr->D))
				{
				
if (DEBUG >= BASIC)
{
	debug << "Counterexample found for SYB(a',b',c')" << endl;	
}
					cout << "Counterexample found for SYB(a',b',c')" << endl;
					exit(0);
				}
			}
						
			lptr++;
		}

		if (S_YANG_BAXTER_SEARCH)
			cout << "All biracks in calaculated list satisfy FYB(a',b',c')" << endl;					
		else
			cout << "All biracks in calaculated list satisfy SYB(a',b',c')" << endl;					
	}


	/* Write out the final calculated list to the output file	*/

if (DEBUG >= INTERMEDIATE)
{
	if (DISPLAY_RELATED_QUANDLE)
	{
		debug << "Checking for related quandles from compare_list:" << endl;
		lptr = compare_list.begin();
		int counter = 0;
		while (lptr != compare_list.end())
		{
			debug << "  " << ++counter << ".";
			debug << " U = ";
			debug_display (lptr->U);
			debug << " D = ";
			debug_display (lptr->D);
			debug << endl;
			lptr++;
		}
	}
}

	int counter = 1;
	
	lptr = calculated_list.begin();
	while (lptr != calculated_list.end())
	{

if (DEBUG >= INTERMEDIATE)
{
	debug << "Writing item " << counter << " to output U = ";
	debug_display (lptr->U);
	debug << " D = ";
	debug_display (lptr->D);
	debug << endl;
}
		/* We want to ensure that the only case where U is the identity mapping (y^x = y, column j all j)
		   is where D is also the identity.  We may do this by checking whether U is the identity and if so
		   flipping U and D.  Assume we should flip then check what we think is the new D matrix to see whether
		   we should flip back.
		*/
		matrix<int> UU = lptr->D;
		matrix<int> DD = lptr->U;
			
		bool flip_UD = true;
		for (int i=0; i< n && flip_UD; i++)
		for (int j=0; j< n && flip_UD ; j++)
		{
			if (DD[i][j] != j)
			{
				//flip back U and D
				UU = lptr->U;
				DD = lptr->D;
				flip_UD = false;
			}
		}					
if (DEBUG >= INTERMEDIATE)
	debug << "Flip U and D = " << (flip_UD? "true": "false") << endl;
			
		if (TeX_OUTPUT_DISPLAY)
		{
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
				output << "$";
			else
				output << "$$";

			if (DOODLE_SEARCH)
				output << "DS^";			
			else if (RACK_SEARCH)
				output << "R^";
			else if (QUANDLE_SEARCH)
				output << "Q^";
			else if (BIRACK_SEARCH)
				output << "BR^";
			else
				output << "BQ^";
			
			output << n << "_{" << counter++ << "}";

			if (DISPLAY_RELATED_QUANDLE)
			{
				cout << "w" << flush;
				
				/* check whether this list item is quandle-related and if so, find the quandle
				   in the compare_list (see above comments under QUANDLE_RELATIVES_SEARCH
				*/
				output << "(";
				if (preserved_diagonal(DD) && Yang_Baxter_satisfied(DD,Identity,true))
				{
if (DEBUG >= INTERMEDIATE)
	debug << "UU is quandle related, checking list for related quandle" << endl;

					int position = check_list(DD,Identity,compare_list,false,false);

if (DEBUG >= INTERMEDIATE)
	debug << "check_list returned quandle position " << position << endl;

					output << "Q^" << n << "_{" << position << "}";
				}
				else 
					output << "-";
				output << ",";

				if (preserved_diagonal(UU) && Yang_Baxter_satisfied(UU,Identity,true))
				{
if (DEBUG >= INTERMEDIATE)
	debug << "DD is quandle related, checking list for related quandle" << endl;

					int position = check_list(UU,Identity,compare_list,false,false);

if (DEBUG >= INTERMEDIATE)
	debug << "check_list returned quandle position " << position << endl;

					output << "Q^" << n << "_{" << position << "}";
				}
				else 
					output << "-";
				output << ")";
			}

			output << "\\quad U=";

			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				output << "(";
				display_as_perms (UU,output,TeX_FORMAT);
//				output << ")";
			}
			else
				display (UU, output,TeX_FORMAT);
		
			output << "\\quad D=";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				output << "(";
				display_as_perms (DD,output,TeX_FORMAT);
//				output << ")";
			}
			else
				display (DD, output,TeX_FORMAT);

			if ( !CONVERT_MAPS_s_TO_f && !CONVERT_MAPS_f_TO_s)
			{
				output << "\\quad ";
			
				output << "\\hbox{order " << switch_order(UU,DD);
	
				output << "} \\quad ";
	
				if (UU == DD)
					output << "S, ";
			
				if (!RACK_SEARCH && !BIRACK_SEARCH && !QUANDLE_SEARCH)
				{
					bool pseudo_quandle_U = equal_rows(UU);
					bool pseudo_quandle_D = equal_rows(DD);
			
					if (pseudo_quandle_U && pseudo_quandle_D)
						output << "DPQ, ";
					else if (pseudo_quandle_U || pseudo_quandle_D)
						output << "PQ, ";	
				}
			
				int u = constant_columns(UU);
				int v = constant_columns(DD);
	
				output << "c_1 = " << u+v << ", c_2 = " << abs(u-v);
			}

			if (DISPLAY_ROWS_AS_PERMUTATIONS)
				output << "$\\hfil\\break" << endl;
			else
				output << "$$" << endl;
		}
		else if (BRAID_OUTPUT_DISPLAY)
		{
			output << "; ";
			if (DOODLE_SEARCH)
				output << "DS^";			
			else if (RACK_SEARCH)
				output << "R^";
			else if (QUANDLE_SEARCH)
				output << "Q^";
			else if (BIRACK_SEARCH)
				output << "BR^";
			else
				output << "BQ^";
			
			output << n << "_{" << counter++ << "}";
			output << endl;

			if (DOODLE_SEARCH)
				output << "S=DT ";
			else
				output << "S=FT ";

			display(UU,output,BRAID_PROGRAMME_FORMAT);
			output << ",";
			display(DD,output,BRAID_PROGRAMME_FORMAT); 
			output << endl;
		}		
		else 
		{

			if (DOODLE_SEARCH)
				output << "DS^";			
			else if (RACK_SEARCH)
				output << "R^";
			else if (QUANDLE_SEARCH)
				output << "Q^";
			else if (BIRACK_SEARCH)
				output << "BR^";
			else
				output << "BQ^";
			
			output << n << "_{" << counter++ << "} ";




			output << "U = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
				display_as_perms (UU,output);
			else
				display (UU, output);
			output << "  D = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
				display_as_perms (DD,output);
			else
				display (DD, output);
			output << " " << endl;	
		}

		lptr++;

	}

	if (DISPLAY_RELATED_QUANDLE)
	{
		/* we've been putting out 'w' characters, so need a new line */
		cout << "\n" << endl;
	}

	output.flush();
	output.close();

if (DEBUG >= BASIC)
{
    debug.flush();
    debug.close();
}

} /* end of main */

bool Yang_Baxter_satisfied (matrix<int>& U, matrix<int>& D,bool rack_condition)
{
	if (S_YANG_BAXTER_SEARCH)
		return S_Yang_Baxter_rules(U,D,rack_condition);
	else
		return F_Yang_Baxter_rules(U,D,rack_condition);
}


/* If the boolean rack_condition is false (the default) the S_Yang_Baxter_rules function checks 
   the up down and rule of five conditions for a birack.  If rack_condition is true, it only
   tests the up condition for a rack.
*/

bool S_Yang_Baxter_rules (matrix<int>& U, matrix<int>& D,bool rack_condition)
{

if (DEBUG >= BASIC)
	debug << "  S-Yang-Baxter rules: " << endl;

	vector<int> rof(3); //rule of five
	for (int i =0; i< n; i++)
	{
		rof[0]=i;
		for (int j =0; j< n; j++)
		{
			rof[1]=j;
			for (int k =0; k < n; k++)
			{
				rof[2]=k;
				
if (DEBUG >= DETAIL)
	debug << "    rof: " << rof[0] << " " << rof[1] << " " << rof[2] << endl;

				int& a = rof[0];
				int& b = rof[1];
				int& c = rof[2];
					
				/* Up interchange */
				int& a_up_b = U[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_b = " << a_up_b <<endl;

				int& b_up_c = U[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_c = " << b_up_c <<endl;

				int& a_up_b__up_c = U[c][a_up_b];
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c = " << a_up_b__up_c <<endl;

				if (rack_condition)
				{
					/* a^bc = a^{cb^c} */
					int& a_up_c = U[c][a];						
if (DEBUG >= DETAIL)
	debug << "      a_up_c = " << a_up_c <<endl;

					if (a_up_b__up_c != U[b_up_c][a_up_c])
					{
if (DEBUG >= DETAIL)
	debug << "      rack up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c == U[b_up_c][a_up_c]\n" << endl;
					}
				}
				else 
				{
					int& c_down_b = D[b][c];					
if (DEBUG >= DETAIL)
	debug << "      c_down_b = " << c_down_b <<endl;

					int& a_up__c_down_b = U[c_down_b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up__c_down_b = " << a_up__c_down_b <<endl;
					
					if (a_up_b__up_c != U[b_up_c][a_up__c_down_b])
					{
if (DEBUG >= DETAIL)
	debug << "      up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c == U[b_up_c][a_up__c_down_b]\n" << endl;
					}
				}

				if (!rack_condition)
				{
					/* Down interchange */
					int& c_up_b = U[b][c];
if (DEBUG >= DETAIL)
	debug << "      c_up_b = " << c_up_b <<endl;

					int& b_down_c = D[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_down_c = " << b_down_c <<endl;

					int& a_down_b = D[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_down_b = " << a_down_b <<endl;

					int& a_down_b__down_c = D[c][a_down_b];
if (DEBUG >= DETAIL)
	debug << "      a_down_b__down_c = " << a_down_b__down_c <<endl;

					int& a_down__c_up_b = D[c_up_b][a];
if (DEBUG >= DETAIL)
	debug << "      a_down__c_up_b = " << a_down__c_up_b <<endl;
					
					if (a_down_b__down_c != D[b_down_c][a_down__c_up_b])
					{
if (DEBUG >= DETAIL)
	debug << "      down interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      a_down_b__down_c == D[b_down_c][a_down__c_up_b]\n" << endl;
					}
					
					/* Rule of five */
					int& b_up_a = U[a][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_a = " << b_up_a <<endl;
	
					int& a_up_c = U[c][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_c = " << a_up_c <<endl;
	
					int& c_down_a = D[a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down_a = " << c_down_a <<endl;
	
					int& b_up__c_down_a = U[c_down_a][b];
if (DEBUG >= DETAIL)
	debug << "      b_up__c_down_a = " << b_up__c_down_a <<endl;
		
					int& c_down__b_up_a = D[b_up_a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down__b_up_a = " << c_down__b_up_a <<endl;
	
					int& a_down_b__up___c_down__b_up_a = U[c_down__b_up_a][a_down_b];
if (DEBUG >= DETAIL)
	debug << "      a_down_b__up___c_down__b_up_a = " << a_down_b__up___c_down__b_up_a <<endl;
	
					
					if (a_down_b__up___c_down__b_up_a != D[b_up__c_down_a][a_up_c])
					{
if (DEBUG >= DETAIL)
	debug << "      rule of five condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "  fails" << endl;
	
						return false;
					}	
				}
			}
		}
	}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;

	return true;
}


/* F_Yang_Baxter_rules checks the three conditions necessary for 
   the boundary map to be consistent for a Reidemeister III move.
*/
bool F_Yang_Baxter_rules (matrix<int>& U, matrix<int>& D,bool rack_condition)
{

if (DEBUG >= BASIC)
	debug << "  F-Yang-Baxter rules: " << endl;

	vector<int> rof(3); //rule of four
	
	for (int i =0; i< n; i++)
	{
		rof[0]=i;
		for (int j =0; j< n; j++)
		{
			rof[1]=j;
			for (int k =0; k < n; k++)
			{
				rof[2]=k;
				
if (DEBUG >= DETAIL)
	debug << "    rof: " << rof[0] << " " << rof[1] << " " << rof[2] << endl;

				int& a = rof[0];
				int& b = rof[1];
				int& c = rof[2];
					
				/* Up interchange */
				int& a_up_b = U[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_b = " << a_up_b <<endl;

				int& a_up_c = U[c][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_c = " << a_up_c <<endl;

				int& b_up_c = U[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_c = " << b_up_c <<endl;

				int& c_down_b = D[b][c];					
if (DEBUG >= DETAIL)
	debug << "      c_down_b = " << c_down_b <<endl;

				if (rack_condition)
				{
					/* a^bc = a^{cb^c} */
					
					int& a_up_b__up_c = U[c][a_up_b];
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c = " << a_up_b__up_c <<endl;

					if (a_up_b__up_c != U[b_up_c][a_up_c])
					{
if (DEBUG >= DETAIL)
	debug << "      rack up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c == U[b_up_c][a_up_c]\n" << endl;
					}
				}
				else 
				{
					/* a^{bc_b} = a^{cb^c} */
																		
					int& a_up_b__up__c_down_b = U[c_down_b][a_up_b];
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up__c_down_b = " << a_up_b__up__c_down_b <<endl;

					
					if (a_up_b__up__c_down_b != U[b_up_c][a_up_c])
					{
if (DEBUG >= DETAIL)
	debug << "      up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up__c_down_b == U[b_up_c][a_up_c]\n" << endl;
					}
				}
				
				if (!rack_condition)
				{
					/* Down interchange c_{ba^b} = c_{ab_a}*/

					int& c_down_a = D[a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down_a = " << c_down_a <<endl;

					int& b_down_a = D[a][b];
if (DEBUG >= DETAIL)
	debug << "      b_down_a = " << b_down_a <<endl;

					int& c_down_b__down__a_up_b = D[a_up_b][c_down_b];
if (DEBUG >= DETAIL)
	debug << "      c_down_b__down__a_up_b = " << c_down_b__down__a_up_b <<endl;

					
					if (c_down_b__down__a_up_b != D[b_down_a][c_down_a])
					{
if (DEBUG >= DETAIL)
	debug << "      down interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
						return false;
					}
					else
					{
if (DEBUG >= DETAIL)
	debug << "      c_down_b__down__a_up_b == D[b_down_a][c_down_a]\n" << endl;
					}
					
					/* Rule of four b_a^{c_a} = b^c_{a^c}*/
				
					int& b_down_a__up__c_down_a = U[c_down_a][b_down_a];
if (DEBUG >= DETAIL)
	debug << "      b_down_a__up__c_down_a = " << b_down_a__up__c_down_a <<endl;
	
					
					if (b_down_a__up__c_down_a != D[a_up_c][b_up_c])
					{
if (DEBUG >= DETAIL)
	debug << "      rule of four condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "  fails" << endl;
	
						return false;
					}	
					else
					{
if (DEBUG >= DETAIL)
	debug << "      b_down_a__up__c_down_a == D[a_up_c][b_up_c]\n" << endl;
					}
				}
			}
		}
	}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;

	return true;
}

bool biquandle_test(matrix<int>& U, matrix<int>& D)
{
	if (S_YANG_BAXTER_SEARCH)
	{
		/* Writing \Delta(x) = x^{x^{-1}} the biquandle condition requires that x_{\Delta(x)} = \Delta(x) for all x.

		   \Delta(x) = i if U_{xi} = x, and the biquandle condition requires that x_i=i; that is, D_{ix} = i.   
		*/
	
if (DEBUG >= BASIC)
	debug << "  S-Yang-Baxter biquandle condition: " << endl;

		for (int x =0; x< n; x++)
		{

if (DEBUG >= DETAIL)
	debug << "    x = " << x << endl;

			int i; 
			
			for (i =0; i< n; i++)
			{
				if (U[x][i] == x) 
					break;
			}
		
if (DEBUG >= DETAIL)
	debug << "      Delta(x) = " << i << endl;

			if (D[i][x] != i)
			{
		
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
		
				return false;
			}
		}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;


		return true;
	}
	else
	{
		/* With the F definition of biracks the biquandle condition is that x^x = x_x for all x */
if (DEBUG >= BASIC)
	debug << "  F-Yang-Baxter biquandle condition: " << endl;

		for (int x =0; x< n; x++)
		{

if (DEBUG >= DETAIL)
	debug << "    x = " << x << endl;

if (DEBUG >= DETAIL)
{
	debug << "      x^x = " << U[x][x] << endl;
	debug << "      x_x = " << D[x][x] << endl;
}

			if (U[x][x] != D[x][x]) 
			{
				
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
		
				return false;
			}
		}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;


		return true;
	}
}

void display(matrix<int>& M, ostream& os, int format)
{
	if (format == TeX_FORMAT)
		os << "\\pmatrix{";
		
	for (int i=0; i< n; i++)
	{
		for (int j=0; j< n; j++)
		{
			if (NUMBER_FROM_ZERO || format == BRAID_PROGRAMME_FORMAT)			
				os << M[i][j];
			else
				os << M[i][j]+1;
			
			if (j < n-1)
			{
				if (format == TeX_FORMAT)
					os << " & ";
				else if (format == BRAID_PROGRAMME_FORMAT)
					os << ",";
				else
					os << " ";
			}
		}
			
		if (i < n-1)
		{
			if (format == TeX_FORMAT)
				os << " \\cr ";
			else if (format == BRAID_PROGRAMME_FORMAT)
				os << ",";
			else
				os << "  ";
		}
	}

	if (format == TeX_FORMAT)
		os << "}";
}

void debug_display (matrix<int>& M)
{
	bool number_from_zero = NUMBER_FROM_ZERO;
	NUMBER_FROM_ZERO = true;
	display(M,debug);
	NUMBER_FROM_ZERO = number_from_zero;
}

/* check_list looks for three types of isomorphism: Flip, Obv, and permutation, as well as combinations.
   If U and D or one of their isomorphisms appear in the specified list it returns the corresponding
   position in the list, otherwise it returns zero.  
   
   check_list was originally written for the case where U and D correspond to S, that is U = Us and D = Ds but now
   the same function is called when S_YANG_BAXTER_SEARCH = false.  In this case, check_list is supplied with 
   U=Uf and D=Df and proceeds to use the same isomorphism checks for the F case as it does in the case of S.
     
   The Flip operation interchanges U and D and applies equally to the S and F cases.  For the S case, the Obv 
   isomorphism determines the map S^{-1}, used for negative crossings, it is evaluated by considering a
   Reidemeister II move with a positive crossing followed by a negative crossing and using the following
   representations for S and invS:
     
	        a \   / b^a         a \   / b_{\bar a}
	           \ /                 \ /
                \                   /
               / \                 / \
            b /   \ a_b         b /   \ a^{\bar b}
    
    
   In the sideways development of the theory, the positive and negative maps S and S^{-1} are dual to F and F^*, so one
   cannot apply Reidemeister II moves to F in the same way as we do for S. However, for a positive crossing we can write
   invF = F^{-1} (x,y) = (y^{\bar x}, x_{\bar y}), by analogy with the above description of invS.
   
   
	      y_x \   / x^y            x \   / y
	           \ /                    \ /
                \                      \
               / \                    / \
            x /   \ y     y^{\bar x} /   \ x_{\bar y}
   
   
   In the above diagrams on the left hand side F operates bottom up and on the right hand side invF operates top down.
   Applying the right hand notation to the left hand diagram we have (x^y)^{\bar{y_x}} = x and (y_x)_{\bar{x^y}} = y.  
   This is precisely what we calcualte below for ObvU and ObvD respectively, so that when S_YANG_BAXTER_SEARCH = false, 
   Obv correctly evaluates the inverse map F^{-1}.

   The permutation isomorphism takes an up and down action and for some permutation \sigma, evaluates:
   
		a \wedge b = \sigma({\sigma^{-1}(a)}^{\sigma^{-1}(b)})
		a \vee b =   \sigma({\sigma^{-1}(a)}_{\sigma^{-1}(b)})

   Where we use the term sigma(S) rather than sigma_sharp(S). These permutations define isomporhic switches in the case 
   where ^ and _ represent s^b and s_b and they define isomporphic sideways maps when they represent f^b and f_b.
   
   In both the S and F developments of the theory we have Flip(Obv) = Obv(Flip), so for each permutation (including the 
   identity) we look for equality, Flip, Obv and Flip(Obv) in the list to be checked.  
      
   If the boolean add is true then we add U,D to the list if there is no isomorphic switch already in the list.
   If the boolean swap_trivial_D_action is true, we add U,D to the list if D is the trivial D action, even if 
   there is an isomorphic switch already there, but in this case we remove that isomorphic switch.  The effect is to 
   create a list that includes a representative from the isomorphism class with a trivial D action wherever possible.
   
*/
int check_list( matrix<int>& U, matrix<int>& D, list<Switch>& list_to_check, bool add, bool swap_trivial_D_action)
{
	list<Switch>::iterator lptr=list_to_check.begin();
	
	matrix <int> SigmaU(n,n), SigmaD(n,n);
	matrix <int> ObvU(n,n), ObvD(n,n);
	
	/* if D is not the trivial action, we won't be swaping anything */
	if (!trivial_action(D))
		swap_trivial_D_action = false;
				
	/* the identity permutation will check U and D as provided */
	vector<int> perm(n), inv_perm(n);
	for (int i=0; i< n; i++)
		perm[i]=i;
		
	do
	{
						
		for (int i=0; i< n; i++)
			inv_perm[perm[i]]=i;

if (DEBUG >= INTERMEDIATE)
{
	debug << "    perm isomorphism check with sigma = ";
	for (int i=0; i< n; i++)
		debug << perm[i] << ' ';
	debug << ", inv_sigma = ";
	for (int i=0; i< n; i++)
		debug << inv_perm[i] << ' ';
	debug << endl;
}
							
		/* create Sigma(S) */
		for (int a=0; a< n; a++)
		for (int b=0; b< n; b++)
		{
			int inv_sigma_a__up__inv_sigma_b = U[inv_perm[b]][inv_perm[a]];
			int inv_sigma_a__down__inv_sigma_b = D[inv_perm[b]][inv_perm[a]];
			SigmaU[b][a] = perm[inv_sigma_a__up__inv_sigma_b];
			SigmaD[b][a] = perm[inv_sigma_a__down__inv_sigma_b];
		}
					
if (DEBUG >= INTERMEDIATE)
{
	debug << "    SigmaU = ";
	debug_display (SigmaU);
	debug << " SigmaD = ";
	debug_display (SigmaD);
	debug << endl;
}

		/* calculate Obv(Sigma), we have to create matrices ObvU and ObvD
		   so that the (x,y) entry of ObvU is y^{\bar x} and the (x,y) entry of ObvD is y_{\bar x}.
		   
		   From the Reidemeister II move with a positive crossing followed by a negative crossing

		   we have that (b^a)^{\bar{a_b}}=b and (a_b)_{\bar{b^a}} = a, so that the (a_b, b^a) entry
		   of ObvU is b and the (b^a, a_b) entry of ObvD is a.
		   
		   (see comments above for a description of why this is still correct when S_YANG_BAXTER_SEARCH = false)
		*/
					
		for (int a=0; a< n; a++)
		for (int b=0; b< n; b++)
		{
			int b_up_a = SigmaU[a][b];
			int a_down_b = SigmaD[b][a];
						
			ObvU[a_down_b][b_up_a]=b; // that is (b^a)^{\bar{a_b}} = b
			ObvD[b_up_a][a_down_b]=a; // that is (a_b)_{\bar{b^a}} = a		
		}

if (DEBUG >= INTERMEDIATE)
{
	debug << "    Obv(Sigma(S)):" << endl;
	debug << "    ObvU = ";
	debug_display (ObvU);
	debug << " ObvD = ";
	debug_display (ObvD);
	debug << endl;
}

if (DEBUG >= DETAIL)
{
	matrix<int> checkU(n,n), checkD(n,n);
	
	/* we should have invS(S(a,b)) = (a,b) that is invS(b^a,a_b) = (a,b)
	   and we have invS(a,b) = (b_{\bar a}, a^{\bar b})

      From the Reidemeister II move with a negative crossing followed by a positive crossing
      we have that (a^{\bar b})^{b_{\bar a}}=a and (b_{\bar a})_{a^{\bar b}} = b, so set the 
	  (a,b) entry of checkU to (a^{\bar b})^{b_{\bar a}} and the (a,b) entry of checkD to
	  (b_{\bar a})_{a^{\bar b}} so that row i of check i should be all i and column j of
	  checkD should be all j.
	*/

	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int a_up_barb = ObvU[b][a];
		int b_down_bara = ObvD[a][b];
						
		checkU[a][b] = SigmaU[b_down_bara][a_up_barb];
		checkD[a][b] = SigmaD[a_up_barb][b_down_bara];
	}
	debug << "checkU = " << checkU << " checkD = " << checkD << endl;
	
}

		lptr=list_to_check.begin();
		int position = 1;
		while (lptr != list_to_check.end())
		{
			if (lptr->D == SigmaD && lptr-> U == SigmaU)
			{

if (DEBUG >= BASIC)
{
	debug << "    equality (i.e. Sigma(S)) found with sigma = ";
	for (int i=0; i< n; i++)
		debug << perm[i] << ' ';
	debug << " in list" << endl;
}	

				if (swap_trivial_D_action)
				{
					list_to_check.push_back(Switch(U,D));
					list_to_check.erase(lptr);
if (DEBUG >= BASIC)
	debug << "    replaced with isomorphism class representative having trivial D action" << endl;
				}
				
				return position;
			}	
			
			if (lptr->D == SigmaU && lptr-> U == SigmaD)
			{

if (DEBUG >= BASIC)
	debug << "    Flip found in list" << endl;
	
				if (swap_trivial_D_action)
				{
					list_to_check.push_back(Switch(U,D));
					list_to_check.erase(lptr);
if (DEBUG >= BASIC)
	debug << "    replaced with isomorphism class representative having trivial D action" << endl;
				}
				return position;
			}	
							
			if (!OMIT_INVERSE_SYMMETRY && lptr->D == ObvD && lptr-> U == ObvU)
			{
		
if (DEBUG >= BASIC)
	debug << "    Obv found in list" << endl;

				if (swap_trivial_D_action)
				{
					list_to_check.push_back(Switch(U,D));
					list_to_check.erase(lptr);
if (DEBUG >= BASIC)
	debug << "    replaced with isomorphism class representative having trivial D action" << endl;
				}
				return position;	
			}

			if (!OMIT_INVERSE_SYMMETRY && lptr->D == ObvU && lptr-> U == ObvD)
			{

if (DEBUG >= BASIC)
	debug << "    Flip(Obv) = Obv(Flip) found in list" << endl;

				if (swap_trivial_D_action)
				{
					list_to_check.push_back(Switch(U,D));
					list_to_check.erase(lptr);
if (DEBUG >= BASIC)
	debug << "    replaced with isomorphism class representative having trivial D action" << endl;
				}
				return position;
			}

			lptr++;
			position++;
		}
	} while (!EXCLUDE_PERMUTATIONS && next_permutation(perm.begin(),perm.end()));


	/* We've got a genuinely new biquandle, add it to the list */
	if (add)
	{
		list_to_check.push_back(Switch(U,D));
	
if (DEBUG >= BASIC)
	debug << "  distinct isomorphism class representative, adding to list" << endl;

	}
	else
	{
if (DEBUG >= BASIC)
	debug << "  new isomorphism class, not found in list" << endl;
	}
	
	return 0;
	
}

/* There is a trivial action if each row of either M (either U or D) is 0..n
   so that y->y^x or y->y_x is the identity for all x
*/
bool trivial_action (matrix<int>& M)
{
	for (int i=0; i< n; i++)
	for (int j=0; j< n; j++)
	{
		if (M[i][j] != j)
			return false;
	}

	return true;
}


/* There is a preserved diagonal if for all x x^x = x 
   or x_x = x depending on whether we're checking U or D
*/
bool preserved_diagonal (matrix<int>& M)
{
	for (int i=0; i< n; i++)
	{
		if (M[i][i] != i)
			return false;
	}
	return true;
}

bool equal_rows (matrix<int>& M)
{
	for (int r=1; r< n; r++)
	for (int c=0; c< n; c++)
	{
		if (M[r][c] != M[0][c])
		{
			return false;
		}
	}
	return true;
}

int constant_columns (matrix<int>& M)
{
	int num=n;

	for (int c=0; c< n; c++)
	{
		for (int r=1; r< n; r++)
		{
			if (M[r][c] != M[0][c])
			{
				num--;
				break;
			}
		}
	}
	
	return num;
}

/* power_2 returns true if the switch indicated by the parameter has the property S^2=1
 
   The effect of a switch S with up and down matrices U and D on the pair (x,y) is given by
   S(x,y) = (U[x][y],D[y][x]), since the (x,y) entry of D and U are y_x and y^x respectively
   and S(x,y) = (y^x,x_y).
   
   Therefore power_2 looks at all combinations of (x,y) and checks that S(y^x,x_y) = (x,y)
*/ 
bool power_2(matrix<int>& U, matrix<int>& D)
{
	for (int i=0; i < n; i++)
	for (int j=0; j < n; j++)
	{		
		if (U[ U[i][j] ][ D[j][i] ] != i)
			return false;

		if (D[ D[j][i] ][ U[i][j] ] != j)
			return false;
	}
	return true;
}

/* switch_order calculates the least k such that S^k(x,y) = (x,y).
   We shall calculate up and down matrices U^k and D^k for S^k and will have
   S^k(x,y)=(x,y) if the (x,y) entry of U^k is x and the (y,x) entry of D^k is y,
   i.e the row i of U^k and D^k are both constant with entries equal to i.
   
   We know that the order of S divides n^2 since S is a permutation of XxX
*/
int switch_order(matrix<int>& U, matrix<int>& D)
{
	matrix<int> Uk(U),Dk(D);
	
	for (int k=1; k <= n*n; k++)
	{
		/* test Uk and Dk for the identity */
		bool identity = true;
		
		for (int i=0; i<n && identity==true; i++)
		for (int j=0; j<n; j++)
		{
			if (Uk[i][j] != i)
			{
				identity = false;
				break;
			}
		}
		
		if (identity)
		{
			for (int i=0; i<n && identity==true; i++)
			for (int j=0; j<n; j++)
			{
				if (Dk[i][j] != i)
				{
					identity = false;
					break;
				}
			}
		}
		
		if (identity)
			return k;
		else
		{
			/* calculate Uk and Dk for the next value of k.  We iterate using
			   S(x,y)=(y^x,x_y) = (U[x][y], D[y][x]) by writing
			   S^k(x,y) = (Uk[x][y],Dk[y][x]) and then we have
			   S^{k+1}(x,y) = S(S^k(x,y)) 
			                = S(Uk[x][y],Dk[y][x])
							= S((Dk[y][x])^(Uk[x][y]),(Uk[x][y])_(Dk[y][x]))
							= ( U[Uk[x][y]][Dk[y][x]], D[Dk[y][x]][Uk[x][y]] ) 
							
			   since the (x,y) entry of U and D are y^x and y_x respectively
			   
			   Thus, 			   
			   
			   U^{k+1}[x][y] = U[Uk[x][y]][Dk[y][x]]
			   D^{k+1}[y][x] = D[Dk[y][x]][Uk[x][y]]
			   
			*/
			matrix<int> locU(n,n),locD(n,n);

			for (int i=0; i<n; i++)
			for (int j=0; j<n; j++)
			{
				locU[i][j] = U[Uk[i][j]][Dk[j][i]];
				locD[j][i] = D[Dk[j][i]][Uk[i][j]];
			}
			
			Uk = locU;
			Dk = locD;
		}
	}
	
	return -1; // error condition, we shouldn't ever get here
}

bool rows_are_permutations (matrix<int>&M)
{
	int n = M.numrows();
	
	for (int i=0; i<n; i++)
	{
		valarray<int> check_perm(0,n);
		for (int j=0; j< n; j++)
		{
			if (check_perm[M[i][j]])
				return false; // not a permutation
			else
				check_perm[M[i][j]] = 1;
		}				
	}
	return true;
}

void display_as_perms (matrix<int>& M, ostream& os, int format)
{
	/* we'll record which elements of the row remain to be written via row_flags */
	int row_flags[n];

	bool identity = true;
	/* first check to see if the matrix is the identity */
	for (int i=0; i< n && identity ; i++)
	{
		for (int j=0; j<n; j++)
			row_flags[j] = 1;

		bool found;
		int start;
	
		do
		{
			found = false;
			int j;
	
			/* look for another starting place in row_flags */
			for (j=0; j<n; j++)
			{
				if (row_flags[j])
				{
					start = j;
					row_flags[j] = 0;
					found = true;
					break;
				}
			}

			if (found)
			{
				if (M[i][j] != start)
				{
					identity = false;
					break;
				}
			}
		} while (found);
	}
	
	if (identity)
	{
		if (format == TeX_FORMAT)
			os << "\\iota";
		else
			os << "i";
	}
	else
	{
		/* we have to print out the matrix row by row, 
		   some of which may be the identity permutation
		*/
		os << "(";

		for (int i=0; i< n; i++)
		{
			for (int j=0; j<n; j++)
				row_flags[j] = 1;

			bool found;
			int start;
	
			identity = true;
			do
			{
				found = false;
				int j;
	
				/* look for another starting place in row_flags */
				for (j=0; j<n; j++)
				{
					if (row_flags[j])
					{
						start = j;
						row_flags[j] = 0;
						found = true;
						break;
					}
				}

				if (found)
				{
					if (M[i][j] != start)
					{
						identity = false;
						os << '(';
						os << j + (NUMBER_FROM_ZERO? 0:1);
			
						do
						{
							os << ' ';
							os << M[i][j] + (NUMBER_FROM_ZERO? 0:1);
								row_flags[M[i][j]] = 0;
								j = M[i][j];
						} while (M[i][j] != start);
					
						os << ')';
					}
				}
			} while (found);
		
			if (identity)
			{
				if (format == TeX_FORMAT)
					os << "\\iota";
				else
					os << "i";
			}
		
			if (i<n-1)
				os << " , ";
		}

		os << ")";
	}
}

/* The function vertex_span counts the number of elements in X={1,...,n}, where n is the size of U
   (that is U is an n x n integer matrix), that can be reached by the action of the inner group of U.
   The inner group of U is that generated by the permutations that are the rows of U.  If all the 
   elements may be reached from 0, the action of U is transitive, and U is indecomposable.  If the
   span of the action is not the entire set X, then U is not transitive and is therefore decomposable.
   
   We check the span of the action by building up a list of vertices that may be reached by the 
   permutations of U.  We initialize the list with 0 and then repreatedly take the next item of 
   the list and apply each permutation to it, checking to see if it takes us to a new element.  If it
   does, we add that element to the end of the list.  When we reach the end of the list, its size
   indicates the number of vertices that may be reached from 1, i.e. is the span of the action (strictly 
   the span of the action from 1).
*/
unsigned int action_span (matrix<int>& U)
{
	int initial_element = 0;
	int n = U.numcols();
	
	list<int> element_list;
	element_list.push_back(initial_element);
	list<int>::iterator lptr = element_list.begin();

	/* we record each visited element in element_flag */
	vector<int> element_flag(n,0);
	element_flag[initial_element] = 1;
	
if (DEBUG >= DETAIL)
{
    debug << "action_span: checking action of ";
	debug_display (U);
	debug << "\naction_span: starting from crossing " << initial_element << endl;
}

	while (lptr != element_list.end())
	{
		/* check which elements may be reached from *lptr, element_flag[i] indicates 
		   whether we already have element i on the list.  Add any new elements
	    */
if (DEBUG >= DETAIL)
    debug << "action_span:   from crossing " << *lptr << endl;

		
		for (int i=0; i<n; i++)
		{
			/* The (x,y) entry of U is y^x and we wish to evaluate *lptr^i, since we
			   action of i on *lptr.  This is given by the entry U[i][*lptr]
			*/
			int element = U[i][*lptr];

if (DEBUG >= DETAIL)
    debug << "action_span:     U[" << i << "][] takes " << *lptr << " U[" << i << "][" << *lptr << "] = " << element;

		    if(element_flag[element] == 0)
		    {
				element_list.push_back(element);
				element_flag[element] = 1;

if (DEBUG >= DETAIL)
    debug << ", adding " << element << " to element_list" << endl;
			}
			else
			{
if (DEBUG >= DETAIL)
    debug << ", which is already on the list " << endl;
			}
		}
		
		lptr++;
	}

if (DEBUG >= DETAIL)
    debug << "action_span: span from element " << initial_element << " = " << element_list.size() << endl;
    
	return element_list.size();
}


/* The function S_sideways_map_bottom_up was inlcuded to test the relation between
   the maps F and S.

   The S sideways map bottom-up is the map 
   
	S_{-}^{+} (a,b) = (b_{a^{-1}}, a^{b_{a^{-1}}})
	
	where the up and down operations on the rhs are g operations and
	where b -> b_{a^{-1}} is the inverse of the down operation b_a.
	
	This map describes how the pair (a,b) at the "bottom" of a positive
	crossing are mapped to the pair at the "top" of the crossing, sideways to 
	the map S.
*/
Switch S_sideways_map_bottom_up(matrix<int>& old_U,matrix<int> old_D)
{
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_sideways_map_bottom_up: old_U:" << endl;
	debug_display (old_U);
	debug << endl;
	debug << "S_sideways_map_bottom_up: old_D:" << endl;
	debug_display (old_D);
	debug << endl;
}

	matrix<int> new_U(n,n);
	matrix<int> new_D(n,n);
	
	matrix<int> inv_D(n,n);
	
	for (int i=0; i< n; i++)
	for (int j=0; j< n; j++)
		inv_D[i][old_D[i][j]] = j; // inverts the map D_i given by the row D[i][]

if (DEBUG >= INTERMEDIATE)
{
	debug << "S_sideways_map_bottom_up: inv_D:" << endl;
	debug_display (inv_D);
	debug << endl;
}
	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int x = inv_D[a][b];  // D_a^{-1}(b)

if (DEBUG >= DETAIL)
	debug << "S_sideways_map_bottom_up:   a = " << a << " b = " << b << " x = " << x << endl;

		new_D[a][b] = x;
		new_U[b][a] = old_U[x][a];
	}

if (DEBUG >= INTERMEDIATE)
{
	debug << "S_sideways_map_bottom_up: new_U:" << endl;
	debug_display (new_U);
	debug << endl;
	debug << "S_sideways_map_bottom_up: new_D:" << endl;
	debug_display (new_D);
	debug << endl;
}
	
	return Switch(new_U, new_D);
}

/* The function F_sideways_map_left_right was inlcuded to test the relation between
   the maps F and S.

   The F sideways map left-right is the map 
   
	F_{-}^{+} (a,b) = (b^{a_{b^{-1}}},a_{b^{-1}})
	
	where the up and down operations on the rhs are f operations and
	where b -> b_{a^{-1}} is the inverse of the down operation b_a.
	
	This map describes how the pair (a,b) at the "left" of a positive
	crossing are mapped to the pair at the "right" of the crossing, sideways to the map F.
*/
Switch F_sideways_map_left_right(matrix<int>& old_U,matrix<int> old_D)
{
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_sideways_map_left_right: old_U:" << endl;
	debug_display (old_U);
	debug << endl;
	debug << "F_sideways_map_left_right: old_D:" << endl;
	debug_display (old_D);
	debug << endl;
}

	matrix<int> new_U(n,n);
	matrix<int> new_D(n,n);
	
	matrix<int> inv_D(n,n);
	
	for (int i=0; i< n; i++)
	for (int j=0; j< n; j++)
		inv_D[i][old_D[i][j]] = j; // inverts the map D_i given by the row D[i][]

if (DEBUG >= INTERMEDIATE)
{
	debug << "F_sideways_map_left_right: inv_D:" << endl;
	debug_display (inv_D);
	debug << endl;
}
	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int x = inv_D[b][a];  // D_b^{-1}(a)

if (DEBUG >= DETAIL)
	debug << "F_sideways_map_left_right:   a = " << a << " b = " << b << " x = " << x << endl;

		new_D[b][a] = x;
		new_U[a][b] = old_U[x][b];
	}

if (DEBUG >= INTERMEDIATE)
{
	debug << "F_sideways_map_left_right: new_U:" << endl;
	debug_display (new_U);
	debug << endl;
	debug << "F_sideways_map_left_right: new_D:" << endl;
	debug_display (new_D);
	debug << endl;
}
	
	return Switch(new_U, new_D);
}

/* F_Yang_Baxter_rule_check takes a pair of matrices describing g up and down maps
   and evaluates the corresponding sideways map S_{-}^{+}.  It then checks the 
   F-Yang-Baxter rules are satisfied for each a',b',c' where
   a' = a, b' = b_a  = s_a(b), c' = c_{a^b b_a} = s_{b_a}(s_{a^b}(c)) = s_{s_a(b)}(s_{s^b(a)}(c))
*/
bool F_Yang_Baxter_rule_check (matrix<int>& Us, matrix<int>& Ds)
{

if (DEBUG >= BASIC)
	debug << "  F-Yang-Baxter rule check: " << endl;

if (DEBUG >= BASIC)
{
	debug << "\nUs = ";
	debug_display (Us);
	debug << " Ds = ";
	debug_display (Ds);
	debug << endl;
}

	Switch F = S_sideways_map_bottom_up(Us,Ds);

	matrix<int>& U = F.U;
	matrix<int>& D = F.D;

if (DEBUG >= BASIC)
{
	debug << "\nUf = ";
	debug_display (U);
	debug << " Df = ";
	debug_display (D);
	debug << endl;
}

	vector<int> rof(3); //rule of four
	
	for (int i =0; i< n; i++)
	{
		rof[0]=i;
		for (int j =0; j< n; j++)
		{
			rof[1]=j;
			for (int k =0; k < n; k++)
			{
				rof[2]=k;
				
if (DEBUG >= DETAIL)
	debug << "    rof: " << rof[0] << " " << rof[1] << " " << rof[2] << endl;

				int& a_s = rof[0];				
				int& b_s = rof[1];
				int& c_s = rof[2];
				
				// a' = a, b' = b_a, c' = c_{a^b b_a}
				int a = a_s;
				int b = Ds[a_s][b_s];
				int a_s__up__b_s = Us[b_s][a_s];
				int c = Ds[a_s__up__b_s][c_s];
				c = Ds[b][c]; // b = b'


if (DEBUG >= DETAIL)
	debug << "    a = " << a << ", b = " << b << ", c = " << c << endl;
								
				/* Up interchange */
				int& a_up_b = U[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_b = " << a_up_b <<endl;

				int& a_up_c = U[c][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_c = " << a_up_c <<endl;

				int& b_up_c = U[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_c = " << b_up_c <<endl;

				int& c_down_b = D[b][c];					
if (DEBUG >= DETAIL)
	debug << "      c_down_b = " << c_down_b <<endl;

				/* a^{bc_b} = a^{cb^c} */
																		
				int& a_up_b__up__c_down_b = U[c_down_b][a_up_b];
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up__c_down_b = " << a_up_b__up__c_down_b <<endl;

					
				if (a_up_b__up__c_down_b != U[b_up_c][a_up_c])
				{
if (DEBUG >= DETAIL)
	debug << "      up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
					return false;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up__c_down_b == U[b_up_c][a_up_c]\n" << endl;
				}
				
				/* Down interchange c_{ba^b} = c_{ab_a}*/

				int& c_down_a = D[a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down_a = " << c_down_a <<endl;

				int& b_down_a = D[a][b];
if (DEBUG >= DETAIL)
	debug << "      b_down_a = " << b_down_a <<endl;

				int& c_down_b__down__a_up_b = D[a_up_b][c_down_b];
if (DEBUG >= DETAIL)
	debug << "      c_down_b__down__a_up_b = " << c_down_b__down__a_up_b <<endl;

					
				if (c_down_b__down__a_up_b != D[b_down_a][c_down_a])
				{
if (DEBUG >= DETAIL)
	debug << "      down interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
					return false;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "      c_down_b__down__a_up_b == D[b_down_a][c_down_a]\n" << endl;
				}
					
				/* Rule of four b_a^{c_a} = b^c_{a^c}*/
			
				int& b_down_a__up__c_down_a = U[c_down_a][b_down_a];
if (DEBUG >= DETAIL)
	debug << "      b_down_a__up__c_down_a = " << b_down_a__up__c_down_a <<endl;
	
					
				if (b_down_a__up__c_down_a != D[a_up_c][b_up_c])
				{
if (DEBUG >= DETAIL)
	debug << "      rule of four condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "  fails" << endl;
	
					return false;
				}	
				else
				{
if (DEBUG >= DETAIL)
	debug << "      b_down_a__up__c_down_a == D[a_up_c][b_up_c]\n" << endl;
				}
			}
		}
	}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;

	return true;
}

/* S_Yang_Baxter_rule_check takes a pair of matrices describing f up and down maps
   and evaluates the corresponding sideways map F_{-}^{+}.  It then checks the 
   S-Yang-Baxter rules are satisfied for each a',b',c' where
   a' = a, b' = b_a  = f_a(b), c' = c_{ba^b} = f_{a^b}(c_b) = f_{f^b(a)}(f_b(c))
*/
bool S_Yang_Baxter_rule_check (matrix<int>& Uf, matrix<int>& Df)
{

if (DEBUG >= BASIC)
	debug << "  S-Yang-Baxter rule check: " << endl;

if (DEBUG >= BASIC)
{
	debug << "\nUf = ";
	debug_display (Uf);
	debug << " Df = ";
	debug_display (Df);
	debug << endl;
}

	Switch S = F_sideways_map_left_right(Uf,Df);

	matrix<int>& U = S.U;
	matrix<int>& D = S.D;

if (DEBUG >= BASIC)
{
	debug << "\nUs = ";
	debug_display (U);
	debug << " Ds = ";
	debug_display (D);
	debug << endl;
}


	vector<int> rof(3); //rule of five
	for (int i =0; i< n; i++)
	{
		rof[0]=i;
		for (int j =0; j< n; j++)
		{
			rof[1]=j;
			for (int k =0; k < n; k++)
			{
				rof[2]=k;
				
if (DEBUG >= DETAIL)
	debug << "    rof: " << rof[0] << " " << rof[1] << " " << rof[2] << endl;

				int& a_f = rof[0];				
				int& b_f = rof[1];
				int& c_f = rof[2];
				
				// a' = a, b' = b_a, c' = c_{ba^b}
				int a = a_f;
				int b = Df[a_f][b_f];
				int a_f__up__b_f = Uf[b_f][a_f];
				int c = Df[b_f][c_f];
				c = Df[a_f__up__b_f][c];

if (DEBUG >= DETAIL)
	debug << "    a = " << a << ", b = " << b << ", c = " << c << endl;
					
			/* Up interchange */
				int& a_up_b = U[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_b = " << a_up_b <<endl;

				int& b_up_c = U[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_c = " << b_up_c <<endl;

				int& a_up_b__up_c = U[c][a_up_b];
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c = " << a_up_b__up_c <<endl;

				int& c_down_b = D[b][c];					
if (DEBUG >= DETAIL)
	debug << "      c_down_b = " << c_down_b <<endl;

				int& a_up__c_down_b = U[c_down_b][a];
if (DEBUG >= DETAIL)
	debug << "      a_up__c_down_b = " << a_up__c_down_b <<endl;
					
				if (a_up_b__up_c != U[b_up_c][a_up__c_down_b])
				{
if (DEBUG >= DETAIL)
	debug << "      up interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
					return false;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "      a_up_b__up_c == U[b_up_c][a_up__c_down_b]\n" << endl;
				}

				/* Down interchange */
				int& c_up_b = U[b][c];
if (DEBUG >= DETAIL)
	debug << "      c_up_b = " << c_up_b <<endl;

				int& b_down_c = D[c][b];
if (DEBUG >= DETAIL)
	debug << "      b_down_c = " << b_down_c <<endl;

				int& a_down_b = D[b][a];
if (DEBUG >= DETAIL)
	debug << "      a_down_b = " << a_down_b <<endl;

				int& a_down_b__down_c = D[c][a_down_b];
if (DEBUG >= DETAIL)
	debug << "      a_down_b__down_c = " << a_down_b__down_c <<endl;

				int& a_down__c_up_b = D[c_up_b][a];
if (DEBUG >= DETAIL)
	debug << "      a_down__c_up_b = " << a_down__c_up_b <<endl;
					
				if (a_down_b__down_c != D[b_down_c][a_down__c_up_b])
				{
if (DEBUG >= DETAIL)
	debug << "      down interchange condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "    fails" << endl;
	
					return false;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "      a_down_b__down_c == D[b_down_c][a_down__c_up_b]\n" << endl;
				}
					
				/* Rule of five */
				int& b_up_a = U[a][b];
if (DEBUG >= DETAIL)
	debug << "      b_up_a = " << b_up_a <<endl;
	
				int& a_up_c = U[c][a];
if (DEBUG >= DETAIL)
	debug << "      a_up_c = " << a_up_c <<endl;
	
				int& c_down_a = D[a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down_a = " << c_down_a <<endl;
	
				int& b_up__c_down_a = U[c_down_a][b];
if (DEBUG >= DETAIL)
	debug << "      b_up__c_down_a = " << b_up__c_down_a <<endl;
		
				int& c_down__b_up_a = D[b_up_a][c];
if (DEBUG >= DETAIL)
	debug << "      c_down__b_up_a = " << c_down__b_up_a <<endl;
	
				int& a_down_b__up___c_down__b_up_a = U[c_down__b_up_a][a_down_b];
if (DEBUG >= DETAIL)
	debug << "      a_down_b__up___c_down__b_up_a = " << a_down_b__up___c_down__b_up_a <<endl;
	
					
				if (a_down_b__up___c_down__b_up_a != D[b_up__c_down_a][a_up_c])
				{
if (DEBUG >= DETAIL)
	debug << "      rule of five condition fails" << endl;
if (DEBUG >= BASIC)
	debug << "  fails" << endl;
	
					return false;
				}	
			}
		}
	}
	
if (DEBUG >= BASIC)
	debug << "    OK" << endl;

	return true;
}

/* Check that U,D define a permutation of XxX */
bool permutation_of_XxX (matrix<int>& U, matrix<int>& D)
{
if (DEBUG >= BASIC)
	debug << "\npermutation_of_XxX check: " << endl;
	
	for (int k=0; k< n; k++)
	{
		/* for each U[i][j]=k, D[j][i] must be a permutation */
		valarray<int> perm(0,n);
		for (int i=0; i< n; i++)
		for (int j=0; j< n; j++)
		{
			if (U[i][j] == k)
			{
				if (perm[D[j][i]])
				{
if (DEBUG >= BASIC)
{
	debug << "    D[j][i] is not a permutation for U[i][j]=" << k << endl;
	debug << "    fails" << endl;
}					
					return false;
				}
				else
					perm[D[j][i]] = 1;
			}
		}
		
	}
			
if (DEBUG >= BASIC)
	debug << "    OK" << endl;
	
	return true;
}
