/************************************************************************
  dominant pairs calculates lists of essential virtual pairs, essential 
  welded pairs and virtual invariant pairs from an input file containing 
  either biquandles or quandle-related biquandles, depending on the order 
  (size) of the biquandles in the input file.
  
  
					   A. Bartholomew
	Original coding April 2009
	Extended to evaluate essential singular pairs February 2014.
	Added essential doodle pairs December 2014

****************************************************************************/
using namespace std;

#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <list>
#include <ctype.h>
#include <stdio.h>
#include <vector>
#include <valarray>

ofstream    debug;

#include <util.h>
#include <matrix.h>

#define OFF						0
#define BASIC					1
#define INTERMEDIATE			2
#define DETAIL 					3

#define PLAIN_FORMAT					0
#define TeX_FORMAT						1
#define BRAID_PROGRAMME_FORMAT 			2
#define BRAID_PROGRAMME_FORMAT_WELDED 	3


unsigned int DEBUG = OFF;

bool SWITCH_SEARCH = false;
bool TeX_OUTPUT_DISPLAY = false;
bool SWITCH_OUTPUT_DISPLAY = false;
bool CHECK_INPUT_FILE = false;
bool BIRACK_SEARCH = false;
bool NO_SYMMETRY_CHECK = false;
bool SINGLE_OUTPUT = false; // links output file name to input file name
bool NUMBER_FROM_ZERO = false;
bool DISPLAY_ROWS_AS_PERMUTATIONS = false;
bool WELD_TO_UNDERSIDE = false;
bool DOODLE_SEARCH = false;

int n; // size of finite field used in search
int num_essential_essential_virtual_pairs = 0;
int num_virtual_invariant_pairs = 0;
int num_essential_welded_pairs = 0;
int num_essential_singular_pairs = 0;
int num_essential_doodle_pairs = 0;

void display(matrix<int>& M, ostream& os, int format=0);
void display_as_perms(matrix<int>& M, ostream& os, int format=0);

struct named_switch
{

	named_switch(matrix<int> u,matrix<int> d,string s):U(u),D(d),name(s){}
	matrix<int> U;
	matrix<int> D;
	string name;
};

bool check_list( matrix<int>& U, matrix<int>& D, list<named_switch>& biquandles, bool add=true);
bool one_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db);
bool two_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db);
bool power_2(list<named_switch>::iterator sptr);
void dominant_test(list<named_switch>::iterator s1, list<named_switch>::iterator s2);
void display(list<pair<named_switch,named_switch> >& L, ostream& os,int format=0);
bool singularly_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db);
bool fully_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db, bool requirement);
bool commuting_switches(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db);

list<pair<named_switch,named_switch> > essential_virtual_pairs;
list<pair<named_switch,named_switch> > virtual_invariant_pairs;
list<pair<named_switch,named_switch> > essential_welded_pairs;
list<pair<named_switch,named_switch> > essential_singular_pairs;
list<pair<named_switch,named_switch> > essential_doodle_pairs;

int main(int argc, char* argv[])
{
	bool options;
	n=3; 
	
	if (argc >= 2)
	{
		/* argc >1 */
		if ( *(argv[1]) == '-')
			options = true;

		if (options)
		{
			char* dptr = strchr(argv[1], '#');
			if (dptr != 0)
			{
				DEBUG = BASIC;
				if (isdigit(*++dptr))
				{
					if (*dptr == '2')
						DEBUG = INTERMEDIATE;
					else if (*dptr == '3')
						DEBUG = DETAIL;
				}
			}

			if (strchr(argv[1], 'n'))
			{
	    		char* c1 = strchr(argv[1],'n')+1;
	    		if (*c1 == '=')
	    		{
					char* c2 = ++c1 + 1;
					while(isdigit(*c2))
			    		c2++;
					get_number(n,c1);
	    		}
			}

			if (strchr(argv[1], 'd'))
				DOODLE_SEARCH = true;

			if (strchr(argv[1], 'B'))
				SWITCH_OUTPUT_DISPLAY = true;

			if (strchr(argv[1], 'P'))
				DISPLAY_ROWS_AS_PERMUTATIONS = true;

			if (strchr(argv[1], 's'))
				SINGLE_OUTPUT = true;

			if (strchr(argv[1], 'T'))
				TeX_OUTPUT_DISPLAY = true;

			if (strchr(argv[1], 'U'))
				WELD_TO_UNDERSIDE = true;

			if (strchr(argv[1], 'Z'))
				NUMBER_FROM_ZERO = true;
		}
	}
	else if (argc == 1 || argc > 3)
	{
		cout << "\n\nUsage: dominant-pairs -n=<n>[#[1|2|3]dBPsTUZ] <input_file1> [<input_file2>]" << endl;
		cout << "      n=<n> specify the size of the field (default 3)" << endl;
		cout << "      #[1|2|3] generate debug output" << endl;
		cout << "        optionally specify level: 1=BASIC, 2=INTERMEDIATE, 3=DETAIL (default BASIC)" << endl;
		cout << "      d doodle search, requires S-candidates in <input_file1> and T-candidates in <input_file2>" << endl;
		cout << "      B display output formatted as input switches for the braid programme" << endl;
		cout << "      P display rows as permutations" << endl;
		cout << "      s creates a single output file rather than separate ones" << endl;
		cout << "      T display matrix output in TeX format" << endl;
		cout << "      U swap the weld conditions, so weld crossings are welded to the underside of the plane" << endl;
		cout << "      Z input and output X_n={0,...,n-1} rather than X_n={1,...,n}" << endl;
		cout << "      <input_file1> specify the file containing the source biquandles (S-candidates for a doodle search)" << endl;
		cout << "      <input_file2> specify the file containing the T-candidates for a doodle search" << endl;
		exit(0);
	}

if (DEBUG >= BASIC)
{
    /* establish a debug file */
    debug.open ("dominant.dbg");

    if (!debug)
    {
        cout << "\nError opening debug file";
        exit(0);
    }
}

	string input_file1 = (options?argv[2]:argv[1]);
	string output_file = "dominant.out"; // default output filename

	ifstream input1;
	input1.open(input_file1.c_str());
	if (input1)
	{
		matrix<int> U(n,n), D(n,n);
		string next_line;

		/* create a list from <input_file1> */
		list<named_switch> input_list_1;			
		while (getline(input1,next_line))
		{
			if (next_line.length())
			{
				if (next_line.find('U') != string::npos)
				{
					string name;
					if (next_line.find('}') != string::npos)
						name = next_line.substr(2,next_line.find('}')-1);
		
					istringstream iss(next_line);
					char ch;
										
					do
					{
						iss >> ch;	
					} while (!(ch == 'U'));
						
					for (int i=0; i< n; i++)
					for (int j=0; j< n; j++)
					{
						do
						{
							iss >> ch;	
							if (isdigit(ch))
								iss.putback(ch);
						} while (!isdigit(ch));
						iss >> U[i][j];
		
						/* The code is written with X_n={0,...,n-1} so if the
						   input is not in that form adjust it */
						if (!NUMBER_FROM_ZERO)
							U[i][j]--;			
					}
		
					for (int i=0; i< n; i++)
					for (int j=0; j< n; j++)
					{
						do
						{
							iss >> ch;	
							if (isdigit(ch))
								iss.putback(ch);
						} while (!isdigit(ch));
						iss >> D[i][j];
		
						if (!NUMBER_FROM_ZERO)
							D[i][j]--;
					}
		
					input_list_1.push_back(named_switch(U,D,name));
				}
			}
		}

if (DEBUG >= BASIC)
	debug << "read " << input_list_1.size() << " items from " << input_file1 << endl;
	
		if (DOODLE_SEARCH)
		{
			string input_file2 = (options?argv[3]:argv[2]);
		
			ifstream input2;
			input2.open(input_file2.c_str());
			if (input2)
			{
				matrix<int> U(n,n), D(n,n);
				string next_line;
		
				/* create a list from <input_file2> */
				list<named_switch> input_list_2;			
				while (getline(input2,next_line))
				{
					if (next_line.length())
					{
						if (next_line.find('U') != string::npos)
						{
							string name;
							if (next_line.find('}') != string::npos)
								name = next_line.substr(2,next_line.find('}')-1);
				
							istringstream iss(next_line);
							char ch;
												
							do
							{
								iss >> ch;	
							} while (!(ch == 'U'));
								
							for (int i=0; i< n; i++)
							for (int j=0; j< n; j++)
							{
								do
								{
									iss >> ch;	
									if (isdigit(ch))
										iss.putback(ch);
								} while (!isdigit(ch));
								iss >> U[i][j];
				
								/* The code is written with X_n={0,...,n-1} so if the
								   input is not in that form adjust it */
								if (!NUMBER_FROM_ZERO)
									U[i][j]--;			
							}
				
							for (int i=0; i< n; i++)
							for (int j=0; j< n; j++)
							{
								do
								{
									iss >> ch;	
									if (isdigit(ch))
										iss.putback(ch);
								} while (!isdigit(ch));
								iss >> D[i][j];
				
								if (!NUMBER_FROM_ZERO)
									D[i][j]--;
							}
				
							input_list_2.push_back(named_switch(U,D,name));
						}
					}
				}

if (DEBUG >= BASIC)
	debug << "read " << input_list_2.size() << " items from " << input_file2 << endl;
	
				/* Check for essential doodle pairs */			
				list<named_switch>::iterator sptr=input_list_1.begin();
				while (sptr != input_list_1.end())
				{
if (DEBUG >= BASIC)
{
	debug << "sptr -> ";
	debug << "U = ";
	display (sptr->U, debug);
	debug << " D = ";
	display (sptr->D, debug);
	debug << endl;
}					list<named_switch>::iterator tptr=input_list_2.begin();
					while (tptr != input_list_2.end())
					{
if (DEBUG >= BASIC)
{
	debug << "tptr -> ";
	debug << "U = ";
	display (tptr->U, debug);
	debug << " D = ";
	display (tptr->D, debug);
	debug << endl;
}
						/* S and T are an essential doodle pair if T has order 2, T dominates S, S has order 2
						   S does not dominate itself, nor T.  Note that we need check only that S does not 
						   one-dominate T, since S has order 2.  Note also that S is assumed here to be a doodle
						   switch, so we already have S^2 = Id and S does not dominate itself (S-Yang-Baxter fails)
						*/
						if (power_2(tptr))
						{									
if (DEBUG >= BASIC)
	debug << "    T^2 is the identity" << endl;
	
							if (one_dominates(tptr->U,tptr->D,sptr->U,sptr->D))
							{
								if (!one_dominates(sptr->U,sptr->D,tptr->U,tptr->D))
								{
if (DEBUG >= BASIC)
	debug << "    S and T are an essential doodle pair" << endl;

									essential_doodle_pairs.push_back(pair<named_switch,named_switch>(*sptr,*tptr));
									num_essential_doodle_pairs++;
								}
								else
								{
if (DEBUG >= BASIC)
	debug << "    S dominates T" << endl;
								}
								
							}
							else
							{
if (DEBUG >= BASIC)
	debug << "    T does not dominate S" << endl;
							}
						}
						else
						{
if (DEBUG >= BASIC)
	debug << "    T^2 is not the identity" << endl;
						}

						tptr++;
					}
					sptr++;
				}
							
				cout << input_file1  << " and " << input_file2 << " determine " 
				     << num_essential_doodle_pairs << " essential doodle pairs\n";
					 
		
				/* prepare output file */
				ofstream output;
				int format;
				
				if (SWITCH_OUTPUT_DISPLAY)
					format = BRAID_PROGRAMME_FORMAT;
				else if (TeX_OUTPUT_DISPLAY)
					format = TeX_FORMAT;
				else
					format = PLAIN_FORMAT;
			
				if (SINGLE_OUTPUT)
				{
					output.open(output_file.c_str());
		
					if (!output)
					{
						cout << "\nError opening output file";
						exit(0);
					}
		
					output << input_file1  << " and " << input_file2 << " determine " 
						   << num_essential_doodle_pairs << " essential doodle pairs\n";
			    	 
					output << "\nEssential doodle pairs:" << endl;
					display(essential_doodle_pairs,output,format);
			
					output.flush();
					output.close();
				}
				else
				{
					ostringstream oss;
					oss << n << ".txt";
					string trailer = oss.str();
		
					output_file = "doodle-pairs-" + trailer;
					output.open(output_file.c_str());
		
					output << "\nEssential doodle pairs:" << endl;				
					display(essential_doodle_pairs,output,format);
		
					output.flush();
					output.close();	
				}							
			}
			else
			{
				cout << "unable to open the input file " << input_file2 << endl;
			}					
		}
		else
		{
			/* Check for dominant pairs amongst switches in the list */
			list<named_switch>::iterator aptr=input_list_1.begin();
			list<named_switch>::iterator a_penultimate=input_list_1.end();
			a_penultimate--;

//			while (aptr != input_list_1.end()--)
			while (aptr != a_penultimate)
			{
if (DEBUG >= BASIC)
{
	debug << "aptr -> ";
	debug << "U = ";
	display (aptr->U, debug);
	debug << " D = ";
	display (aptr->D, debug);
	debug << endl;
}

				list<named_switch>::iterator bptr=aptr;
				bptr++;
				while (bptr != input_list_1.end())
				{

if (DEBUG >= BASIC)
{
	debug << "  bptr -> ";
	debug << "U = ";
	display (bptr->U, debug);
	debug << " D = ";
	display (bptr->D, debug);
	debug << endl;
}

					/* We have to check the pair of switches indicated 
					   by aptr and bptr both ways round since the conditions
					   are not symmetric
					*/

if (DEBUG >= BASIC)
	debug << "  test S=*aptr T=*bptr" << endl;
	
					dominant_test(aptr,bptr);

if (DEBUG >= BASIC)
	debug << "  test S=*bptr T=*aptr" << endl;
	
					dominant_test(bptr,aptr);						

					bptr++;
				}
				
				aptr++;
			}
	
			cout << input_file1 << " contains:\n" 
			     << "\t" << num_essential_essential_virtual_pairs << " essential virtual pairs\n"
			     << "\t" << num_essential_welded_pairs << " essential welded pairs\n"
			     << "\t" << num_essential_singular_pairs << " essential singular pairs\n" 
			     << "\t" << num_virtual_invariant_pairs << " other virtual invariant pairs\n" << endl;
				 
	
			/* prepare output file */
			ofstream output;
			int format;
			
			if (SWITCH_OUTPUT_DISPLAY)
				format = BRAID_PROGRAMME_FORMAT;
			else if (TeX_OUTPUT_DISPLAY)
				format = TeX_FORMAT;
			else
				format = PLAIN_FORMAT;
		
			if (SINGLE_OUTPUT)
			{
				output.open(output_file.c_str());
	
				if (!output)
				{
					cout << "\nError opening output file";
					exit(0);
				}
	
				output << input_file1 << " contains:\n" 
				     << "\t" << num_essential_essential_virtual_pairs << " essential virtual pairs\n"
			    	 << "\t" << num_essential_welded_pairs << " essential welded pairs\n"
				     << "\t" << num_essential_singular_pairs << " essential singular pairs\n" 
					 << "\t" << num_virtual_invariant_pairs << " other virtual invariant pairs\n" << endl;
			    	 
				output << "\nEssential virtual pairs:" << endl;
				display(essential_virtual_pairs,output,format);
	
				output << "\nVirtual invariant pairs:" << endl;
				display(virtual_invariant_pairs,output,format);
	
				if (format ==BRAID_PROGRAMME_FORMAT)
					format = BRAID_PROGRAMME_FORMAT_WELDED;
	
				output << "\nEssential welded pairs:" << endl;
				display(essential_welded_pairs,output,format);			
	
				output << "\nEssential singular pairs:" << endl;
				display(essential_singular_pairs,output,format);			
	
				output.flush();
				output.close();
			}
			else
			{
				ostringstream oss;
				oss << n << ".txt";
				string trailer = oss.str();
	
				output_file = "virtual-pairs-" + trailer;
				output.open(output_file.c_str());
	
				output << "\nEssential virtual pairs:" << endl;
				display(essential_virtual_pairs,output,format);
	
				output.flush();
				output.close();
	
				output_file = "weld-pairs-" + trailer;
				output.open(output_file.c_str());
	
				output << "\nVirtual invariant pairs:" << endl;
				display(virtual_invariant_pairs,output,format);
	
				output.flush();
				output.close();
	
				output_file = "essential-pairs-" + trailer;
				output.open(output_file.c_str());
	
				if (format == BRAID_PROGRAMME_FORMAT)
					format = BRAID_PROGRAMME_FORMAT_WELDED;
	
				output << "\nEssential welded pairs:" << endl;
				display(essential_welded_pairs,output,format);
	
				output.flush();
				output.close();
	
				output_file = "essential-singular-pairs-" + trailer;
				output.open(output_file.c_str());
	
				if (format == BRAID_PROGRAMME_FORMAT)
					format = BRAID_PROGRAMME_FORMAT_WELDED;
	
				output << "\nEssential singular pairs:" << endl;
				display(essential_singular_pairs,output,format);
	
				output.flush();
				output.close();
			}
		}
	}
	else
		cout << "unable to open the input file " << input_file1 << endl;



if (DEBUG >= BASIC)
{
    debug.flush();
    debug.close();
}

} /* end of main */


void display(matrix<int>& M, ostream& os, int format)
{
	if (format == TeX_FORMAT)
		os << "\\pmatrix{";
		
	for (int i=0; i< n; i++)
	{
		for (int j=0; j< n; j++)
		{
			if (NUMBER_FROM_ZERO || format == BRAID_PROGRAMME_FORMAT)			
				os << M[i][j];
			else
				os << M[i][j]+1;
			
			if (j < n-1)
			{
				if (format == TeX_FORMAT)
					os << " & ";
				else if (format == BRAID_PROGRAMME_FORMAT)
					os << " ";
				else
					os << " ";
			}
		}
			
		if (i < n-1)
		{
			if (format == TeX_FORMAT)
				os << " \\cr ";
			else if (format == BRAID_PROGRAMME_FORMAT)
				os << " ";
			else
				os << "  ";
		}
	}

	if (format == TeX_FORMAT)
		os << "}";
}

void display(list<pair<named_switch,named_switch> >& L, ostream& os, int format)
{
	/* We don't need to distinguish the welded format 
	   when writing out the matrices themselves
	*/
	int display_format = format;
	if (format == BRAID_PROGRAMME_FORMAT_WELDED)
		display_format = BRAID_PROGRAMME_FORMAT; 
	
	list<pair<named_switch,named_switch> >::iterator lptr = L.begin();
	while (lptr != L.end())
	{
		if (format == BRAID_PROGRAMME_FORMAT || format == BRAID_PROGRAMME_FORMAT_WELDED)
		{
			os << "; S = " << lptr->first.name << " T = " << lptr->second.name << endl;

			if (DOODLE_SEARCH)
				os << "S=DT";
			else
				os << "S=FT";
			
			if (format == BRAID_PROGRAMME_FORMAT_WELDED)
				os << "W ";
			else 
				os << " ";
					
			display(lptr->first.U,os,display_format);
			os << " ";
			display(lptr->first.D,os,display_format); 
			os << " ";
			display(lptr->second.U,os,display_format);
			os << " ";
			display(lptr->second.D,os,display_format); 
			os << endl;
		}
		else if (format == TeX_FORMAT)
		{
			os << "\n$S: " << lptr->first.name;
			os << " \\quad U = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->first.U,os,TeX_FORMAT);
//				os << ")";
			}
			else
				display(lptr->first.U,os,format);

			os << " \\quad D = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->first.D,os,TeX_FORMAT);
//				os << ")";
			}
			else
				display(lptr->first.D,os,format);
				
			os << "$\\hfil\\break\n$T: " << lptr->second.name;
			os << " \\quad U = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->second.U,os,TeX_FORMAT);
//				os << ")";
			}
			else
				display(lptr->second.U,os,format); 
				
			os << " \\quad D = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->second.D,os,TeX_FORMAT);
//				os << ")";
			}
			else
				display(lptr->second.D,os,format);
				
			os << "$\\hfil\\break" << endl;
		}
		else
		{
			os << "\nS: " << lptr->first.name;
			os << " U = ";

			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->first.U,os);
//				os << ")";
			}
			else
				display(lptr->first.U,os,format);
			
			os << "  D = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->first.D,os);
//				os << ")";
			}
			else
				display(lptr->first.D,os,format);

			os << " \nT: " << lptr->second.name;
			os << " U = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->second.U,os);
//				os << ")";
			}
			else
				display(lptr->second.U,os,format); 

			os << "  D = ";
			if (DISPLAY_ROWS_AS_PERMUTATIONS)
			{
//				os << "(";
				display_as_perms (lptr->second.D,os);
//				os << ")";
			}
			else
				display(lptr->second.D,os,format);

			os << endl;
		}
		lptr++;
	};
}

/* power_2 returns true if the switch indicated by the parameter has the property S^2=1
 
   The effect of a switch S with up and down matrices U and D on the pair (x,y) is given by
   S(x,y) = (U[x][y],D[y][x]), since the (x,y) entry of D and U are y_x and y^x respectively
   and S(x,y) = (y^x,x_y).
   
   Therefore power_2 looks at all combinations of (x,y) and checks that S(y^x,x_y) = (x,y)
*/ 
bool power_2(list<named_switch>::iterator sptr)
{
	matrix<int>& U = sptr->U;
	matrix<int>& D = sptr->D;
	
	for (int i=0; i < n; i++)
	for (int j=0; j < n; j++)
	{		
		if (U[ U[i][j] ][ D[j][i] ] != i)
			return false;

		if (D[ D[j][i] ][ U[i][j] ] != j)
			return false;
	}
	return true;
}

/* commuting_switches checks that for all x,y B(A(x,y)) = A(B(x,y)) 
   where A(x,y) = (Ua[x][y],Da[y][x]) etc.
*/
bool commuting_switches(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db)
{
	for (int x=0; x < n; x++)
	for (int y=0; y < n; y++)
	{		
		if (Ub[Ua[x][y]][Da[y][x]] != Ua[Ub[x][y]][Db[y][x]] || Db[Da[y][x]][Ua[x][y]] != Da[Db[y][x]][Ub[x][y]])
		{
			return false;
		}
	}
	
	return true;
}


/* With S as the switch indicated by aptr and T as the switch indicated by bptr dominant_test 
   first check whether T has power 2.  
   
   If T does have power 2, we go on to check whether S and T are either an essential virtual pair, 
   an essential welded pair or a virtual invriant pair.  For each of these cases we must have 
   T 1-dominating S (in which case it 2-dominates S as well).  Furthermore:
   
    - If S neither one-dominates, nor two-dominates T we have a essential virtual pair
    - If S one-dominates T but does not two-dominate it, we have an essential welded pair
    - If either S both one-domainates and two-dominates T, or if it two-dominates T but 
      does not one-dominate it, we have a virtal invariant pair

   If T does not have power 2, we check whether S and T are an essential singular pair.
   For this we require that
   
    - T commutes with S
    - S fully dominates T (all R3 cases are satisfied for \epsilon and \eta crossings associated with S and \xi associated with T)
    - T does not fully dominates S (all R3 cases fail for \epsilon and \eta crossings associated with T and \xi associated with S)
    - T does not fully dominate itself (all R3 cases fail for \epsilon, \eta and \xi crossings all associated with T)

*/
void dominant_test(list<named_switch>::iterator aptr, list<named_switch>::iterator bptr)
{
	bool essential_virtual_pair = false;
	bool virtual_invariant_pair = false;
	bool essential_welded_pair = false;
	bool essential_singular_pair = false;
	
	if (power_2(bptr))
	{									
if (DEBUG >= BASIC)
	debug << "    T^2 is the identity" << endl;
		if (one_dominates(bptr->U,bptr->D,aptr->U,aptr->D))
		{
			if ( (!WELD_TO_UNDERSIDE && !two_dominates(aptr->U,aptr->D,bptr->U,bptr->D)) || (WELD_TO_UNDERSIDE && !one_dominates(aptr->U,aptr->D,bptr->U,bptr->D)))
			{

				if ( (!WELD_TO_UNDERSIDE && one_dominates(aptr->U,aptr->D,bptr->U,bptr->D)) || (WELD_TO_UNDERSIDE && two_dominates(aptr->U,aptr->D,bptr->U,bptr->D)))
				{

if (DEBUG >= BASIC)
	debug << "    S and T are an essential welded pair" << endl;
				
					essential_welded_pair = true;
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "    S and T are an essential virtual pair" << endl;

					essential_virtual_pair = true;
				}
			
			}
			else
			{
				
if (DEBUG >= BASIC)
	debug << "    S and T are a virtual invariant pair" << endl;

				virtual_invariant_pair = true;
			}
		}
		else
		{
if (DEBUG >= BASIC)
	debug << "    T does not dominate S" << endl;
		}
	}
	else
	{
if (DEBUG >= BASIC)
	debug << "    T^2 is not the identity" << endl;
	
		/* Look for essential singular pairs */
		if (commuting_switches(aptr->U,aptr->D,bptr->U,bptr->D))
		{
if (DEBUG >= BASIC)
	debug << "    S and T commute" << endl;			

			/* check S dominates T as classical crossings dominate singular crossings*/
			if (singularly_dominates(aptr->U,aptr->D,bptr->U,bptr->D))
			{
				/* check T does not dominates S */
				if (fully_dominates(bptr->U,bptr->D,aptr->U,aptr->D,false))
				{
					/* check T does not dominates itself */
					if (fully_dominates(bptr->U,bptr->D,bptr->U,bptr->D,false))
					{
if (DEBUG >= BASIC)
	debug << "    S and T are an essential singular pair" << endl;			
						essential_singular_pair = true;
					}
					else
					{
if (DEBUG >= BASIC)
	debug << "    T dominates itself for some cases" << endl;			
					}
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "    T dominates S for some cases" << endl;			
				}
			}
			else
			{
if (DEBUG >= BASIC)
	debug << "    S does not singularly dominate T" << endl;			
			}
		}
		else
		{
if (DEBUG >= BASIC)
	debug << "    S and T do not commute" << endl;			
		}
	}
	
	/* push the pair onto the appropriate list */
	if (essential_welded_pair)
	{
		essential_welded_pairs.push_back(pair<named_switch,named_switch>(*aptr,*bptr));
		num_essential_welded_pairs++;
	}
	else if (essential_virtual_pair)
	{
		essential_virtual_pairs.push_back(pair<named_switch,named_switch>(*aptr,*bptr));
		num_essential_essential_virtual_pairs++;
	}
	else if (essential_singular_pair)
	{
		essential_singular_pairs.push_back(pair<named_switch,named_switch>(*aptr,*bptr));
		num_essential_singular_pairs++;
	}
	else if (virtual_invariant_pair)
	{
		virtual_invariant_pairs.push_back(pair<named_switch,named_switch>(*aptr,*bptr));
		num_virtual_invariant_pairs++;
	}
}

/* one_dominates tests whether the switch A, determined by up and down matrices Ua and Da
   (a candidate for S), 1-dominates the switch B, determined by up and down matrices Ub and Db
   (a candidate for T).
   
   The swith A 1-dominates B if 
   
   (A x I) (I x A) (B x 1) = (I x B) (A x I) (I x A)   (*)  
   
   Each case is checked by evaluating the matrix representing the composition on
   the lhs and rhs.

   Writing A(x,y) for the effect of the switch A, we have
   (A x I)(i1,i2,i3)=(A(i1,i2),i3) and (I x A)(i1,i2,i3)=(i1,A(i2,i3)), etc..
   
   Note that we are associating the labels i1, i2 and i3 to the braid strands in a top down 
   manner here, consistent with our convention of writing S(x,y) = (y^x,x_y).
   
   If we write S1 = S x Id, and S2 = Id x S, and similarly for T, this function checks the
   relationship
   
		S1S2T1 = T2S1S2
		
   which corresponds to the conditions imposed on S and T by the allowable weld move in a braid.   
   	   
   The effect of a switch S with up and down matrices U and D on the pair (x,y) is given by
   S(x,y) = (U[x][y],D[y][x]), since the (x,y) entry of U and D are y^x and y_x respectively
   and S(x,y) = (y^x,x_y).

   The function compares the lhs and rhs compositions in (*), the composition for the lhs gives 
   (i1,i2,i3) -> (l1,l2,l3), where (i1,i2,i3)->(j1,j2,j3) under A1, (j1,j2,j3)->(k1,k2,k3) 
   under A2 and (k1,k2,k3)->(l1,l2,l3) under B1.  Similarly for the rhs.
   
   We consider each triple (i1,i2,i3) and evaluate the corresponding (l1,l2,l3) for the lhs
   and (r1,r2,r3) for the rhs and then compare them, if they differ, then the test fails.
*/
bool one_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db)
{
	for (int i=0; i < n; i++)
	for (int j=0; j < n; j++)
	for (int k=0; k < n; k++)
	{
		int l1=i,l2=j,l3=k,temp;
		
		/* evaluate the lhs */
		
		/* (aptr x I)(x,y,z) = (Ua[x][y],Da[y][x],z) */
		temp = Ua[l1][l2];
		l2 = Da[l2][l1];
		l1 = temp;

		/* (I x aptr)(x,y,z) = (x, Ua[y][z],Da[z][y]) */
		temp = Ua[l2][l3];
		l3 = Da[l3][l2];
		l2 = temp;
		
		/* (bptr x I)(x,y,z) = (Ub[x][y],Db[y][x],z) */
		temp = Ub[l1][l2];
		l2 = Db[l2][l1];
		l1 = temp;

		int r1=i,r2=j,r3=k;

		/* evaluate the rhs */

		/* (I x bptr)(x,y,z) = (x, Ub[y][z],Db[z][y]) */
		temp = Ub[r2][r3];
		r3 = Db[r3][r2];
		r2 = temp;

		/* (aptr x I)(x,y,z) = (Ua[x][y],Da[y][x],z) */
		temp = Ua[r1][r2];
		r2 = Da[r2][r1];
		r1 = temp;
		
		/* (I x aptr)(x,y,z) = (x, Ua[y][z],Da[z][y]) */
		temp = Ua[r2][r3];
		r3 = Da[r3][r2];
		r2 = temp;
		
		if (r1 != l1 || r2 != l2 || r3 != l3)
			return false;
	}
	
	return true;
}

/* two_dominates tests whether the switch A, determined by up and down matrices Ua and Da
   (a candidate for S), 2-dominates the switch B, determined by up and down matrices Ub and Db
   (a candidate for T).
   
   The swith A 2-dominates B if 
   
   (I x A) (A x I) (I x B) = (B x I) (I x A) (A x I)   (*)  
    
   Each case is checked by evaluating the matrix representing the composition on
   the lhs and rhs.

   Writing A(x,y) for the effect of the switch A, we have
   (A x I)(i1,i2,i3)=(A(i1,i2),i3) and (I x A)(i1,i2,i3)=(i1,A(i2,i3)), etc..
   
   Note that we are associating the labels i1, i2 and i3 to the braid strands in a top down 
   manner here, consistent with our convention of writing S(x,y) = (y^x,x_y).
   
   If we write S1 = S x Id, and S2 = Id x S, and similarly for T, this function checks the
   relationship
   
		S2S1T2 = T1S2S1
		
   which corresponds to conditions imposed on S and T by the forbidden weld move in a braid.

   The effect of a switch S with up and down matrices U and D on the pair (x,y) is given by
   S(x,y) = (U[x][y],D[y][x]), since the (x,y) entry of U and D are y^x and y_x respectively
   and S(x,y) = (y^x,x_y).

   The function compares the lhs and rhs compositions in (*), the composition for the lhs gives 
   (i1,i2,i3) -> (l1,l2,l3), where (i1,i2,i3)->(j1,j2,j3) under A2, (j1,j2,j3)->(k1,k2,k3) 
   under A1 and (k1,k2,k3)->(l1,l2,l3) under B2.  Similarly for the rhs.
   
   We consider each triple (i1,i2,i3) and evaluate the corresponding (l1,l2,l3) for the lhs
   and (r1,r2,r3) for the rhs and then compare them, if they differ, then the test fails.
*/
bool two_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db)
{
	for (int i=0; i < n; i++)
	for (int j=0; j < n; j++)
	for (int k=0; k < n; k++)
	{
		int l1=i,l2=j,l3=k,temp;
		
		/* evaluate the lhs */
		
		/* (I x aptr)(x,y,z) = (x, Ua[y][z],Da[z][y]) */
		temp = Ua[l2][l3];
		l3 = Da[l3][l2];
		l2 = temp;

		/* (aptr x I)(x,y,z) = (Ua[x][y],Da[y][x],z) */
		temp = Ua[l1][l2];
		l2 = Da[l2][l1];
		l1 = temp;

		/* (I x bptr)(x,y,z) = (x, Ub[y][z],Db[z][y]) */
		temp = Ub[l2][l3];
		l3 = Db[l3][l2];
		l2 = temp;

		int r1=i,r2=j,r3=k;

		/* evaluate the rhs */
		
		/* (bptr x I)(x,y,z) = (Ub[x][y],Db[y][x],z) */
		temp = Ub[r1][r2];
		r2 = Db[r2][r1];
		r1 = temp;
		
		/* (I x aptr)(x,y,z) = (x, Ua[y][z],Da[z][y]) */
		temp = Ua[r2][r3];
		r3 = Da[r3][r2];
		r2 = temp;
		
		/* (aptr x I)(x,y,z) = (Ua[x][y],Da[y][x],z) */
		temp = Ua[r1][r2];
		r2 = Da[r2][r1];
		r1 = temp;

		if (r1 != l1 || r2 != l2 || r3 != l3)
			return false;
	}
	
	return true;
}

/* R3_dominate_case checks the particular case of the generalized R3 condition determined by the 
   switches provided to the function for the \epsilon, \eta and \xi crossings.
   
   Specifically, it checks that 
    
   (A^\epsilon x I) (I x B^\xi) (A^\eta x I) = (I x A^\eta) (B^\xi x I) (I x A^\epsilon) 
    
   where A is the switch corresponding to the \epsilon and \eta crossings, and B is the 
   switch corresponding to the \xi crossing.

   For a switch S the operation at a crossing of positive polarity is S(x,y) = (y^x, x_y)
   and at a crossing of negative polarity it is invS(x,y) = (y_{\bar x}, x^{\bar y}) so (working 
   top down), the top operation is the up operation for a positive crossing and is the (inverse) down
   operation for a negative crossing.  We therefore use Tepsilon and Bepsilon etc. to refer to the 
   operation matrices used at the epsilon crossing to cover both the positive and negative polarity cases.

   Each case is checked by evaluating the matrix representing the composition on
   the lhs and rhs.

   The effect of a switch S with up and down matrices U and D on the pair (x,y) is given by
   S(x,y) = (U[x][y],D[y][x]), since the (x,y) entry of U and D are y^x and y_x respectively
   and S(x,y) = (y^x,x_y).  Similarly the effect of a switch invS with up and down matrices 
   invU and invD on the pair (x,y) is given by invS(x,y) = (invD[x][y],invU[y][x]), since the 
   (x,y) entry of invU and invD are y^{\bar x} and y_{\bar x} respectively and 
   invS(x,y) = (y_{\bar{x}},x^{\bar{y}}).  Thus, once the top and bottom operation matrices are 
   set appropriately, the effect of A^\epsilon on the pair (x,y) is 
   A^\epsilon(x,y) = (Tepsilon[x][y],Bepsilon[y][x]), and similarly for the other crossing types.
   (The key point here is the common relative positions of the x and y variables in all cases.)

   The function compares the lhs and rhs compositions in (*), the composition for the lhs gives 
   (i1,i2,i3) -> (l1,l2,l3), where (i1,i2,i3)->(j1,j2,j3) under A1, (j1,j2,j3)->(k1,k2,k3) 
   under B2 and (k1,k2,k3)->(l1,l2,l3) under A1.  Similarly for the rhs.
   
   We consider each triple (i1,i2,i3) and evaluate the corresponding (l1,l2,l3) for the lhs
   and (r1,r2,r3) for the rhs and then compare them, if they differ, then the test fails.
*/
bool R3_dominate_case(matrix<int>& Tepsilon,matrix<int>& Bepsilon,matrix<int>& Teta,matrix<int>& Beta,matrix<int>& Txi,matrix<int>& Bxi)
{
	for (int i=0; i < n; i++)
	for (int j=0; j < n; j++)
	for (int k=0; k < n; k++)
	{
		int l1=i,l2=j,l3=k,temp;
		
		/* evaluate the lhs */
		
		/* (A^\epsilon x I)(x,y,z) = (Tepsilon[x][y],Bepsilon[y][x],z) */
		temp = Tepsilon[l1][l2];
		l2 = Bepsilon[l2][l1];
		l1 = temp;

		/* (I x B^\xi)(x,y,z) = (x,Txi[y][z],Bxi[z][y]) */
		temp = Txi[l2][l3];
		l3 = Bxi[l3][l2];
		l2 = temp;

		/* (A^\eta x I)(x,y,z) = (Teta[x][y],Beta[y][x],z) */
		temp = Teta[l1][l2];
		l2 = Beta[l2][l1];
		l1 = temp;

		int r1=i,r2=j,r3=k;

		/* evaluate the rhs */
		
		/* (I x A^\eta)(x,y,z) = (x,Teta[y][z],Beta[z][y]) */
		temp = Teta[r2][r3];
		r3 = Beta[r3][r2];
		r2 = temp;
		
		/* (B^\xi x I)(x,y,z) = (Txi[x][y],Bxi[y][x],z) */
		temp = Txi[r1][r2];
		r2 = Bxi[r2][r1];
		r1 = temp;
		
		/* (I x A^\epsilon)(x,y,z) = (x,Tepsilon[y][z],Bepsilon[z][y]) */
		temp = Tepsilon[r2][r3];
		r3 = Bepsilon[r3][r2];
		r2 = temp;

		if (r1 != l1 || r2 != l2 || r3 != l3)
			return false;
	}
	
	return true;
}

/* singularly_dominates carries out the suite of tests corresponding to the generalized R3 move
   R_3(\epsilon, \eta; \xi), involving classical crossings with polarity \epsilon and \eta 
   and a singular crossing of type C_j with polarity \xi.  If A is the switch corresponding to the 
   classical crossings and B is the switch corresponding to the singular crossings, singularly_dominates
   requires that the following four relations (corresponding to the allowable moves) are satisfied:
    
   (A^{-1} x I) (I x B})     (A x I)      = (I x A)      (B x I)      (I x A^{-1})
   (A x I)      (I x B})     (A^{-1} x I) = (I x A^{-1}) (B x I)      (I x A)
   (A^{-1} x I) (I x B^{-1}) (A x I)      = (I x A)      (B^{-1} x I) (I x A^{-1})
   (A x I)      (I x B^{-1}) (A^{-1} x I) = (I x A^{-1}) (B^{-1} x I) (I x A)
    

   That is, when the \epsilon and \eta polarities are different, the relationship is required to hold.
   
   It also requires that that the following four relations (corresponding to the non-allowable moves) are 
   not satisfied:
    
   (A x I)      (I x B})     (A x I)      = (I x A)      (B x I)      (I x A)
   (A^{-1} x I) (I x B})     (A^{-1} x I) = (I x A^{-1}) (B x I)      (I x A^{-1})
   (A x I)      (I x B^{-1}) (A x I)      = (I x A)      (B^{-1} x I) (I x A)
   (A^{-1} x I) (I x B^{-1}) (A^{-1} x I) = (I x A^{-1}) (B^{-1} x I) (I x A^{-1})
    
   That is, when the \epsilon and \eta polarities are the same, the relationship is required not to hold.

   Each case is checked by evaluating the matrix representing the composition on
   the lhs and rhs.

   The function returns true if the above requirements are met and false otherwise.
*/
bool singularly_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db)
{

if (DEBUG >= DETAIL)
{
	debug << "singularly_dominates: classical switch A ";
	debug << "Ua = ";
	display (Ua, debug);
	debug << " Da = ";
	display (Da, debug);
	debug << endl;
	debug << "singularly_dominates: singular switch  B ";
	debug << "Ub = ";
	display (Ua, debug);
	debug << " Db = ";
	display (Da, debug);
	debug << endl;
}

	/* calculate the inverse of A and B.  Given a switch S, invS(x,y) = (y_{\bar{x}},x^{\bar{y}}); 
	   we create matrices invSu and invSd so that the (x,y) entry of invSu is y^{\bar x} and 
	   the (x,y) entry of invSd is y_{\bar x}.

	   From the Reidemeister II move with a positive crossing followed by a negative crossing
	   we have that (b^a)^{\bar{a_b}}=b and (a_b)_{\bar{b^a}} = a, so that the (a_b, b^a) entry
	   of invSu is b and the (b^a, a_b) entry of invSd is a.
	*/
	matrix<int> invUa(n,n), invDa(n,n);	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int b_up_a = Ua[a][b];
		int a_down_b = Ua[b][a];
			
		invUa[a_down_b][b_up_a]=b;
		invDa[b_up_a][a_down_b]=a;		
	}

	matrix<int> invUb(n,n), invDb(n,n);	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int b_up_a = Ub[a][b];
		int a_down_b = Db[b][a];
			
		invUb[a_down_b][b_up_a]=b;
		invDb[b_up_a][a_down_b]=a;		
	}
				
	/* in the following loop clauses the value 0 will be taken to mean positive polarity
	   and the value 1 will be taken to mean negative polarity.
	   
	   For a switch S the operation at a crossing of positive polarity is S(x,y) = (y^x, x_y)
	   and at a crossing of negative polarity it is invS(x,y) = (y_{\bar x}, x^{\bar y}) so (working 
	   top down), the top operation is the up operation for a positive crossing and is the (inverse) down
	   operation for a negative crossing.  We therefore use Tepsilon and Bepsilon etc. to refer to the 
	   top and bottom operation matrices used at the epsilon crossing to cover both the positive and 
	   negative polarity cases.
	*/

	for (int i=0; i<2; i++) // \epsilon
	{
		matrix<int>* Tepsilon;
		matrix<int>* Bepsilon;
		if (i == 0)
		{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: i = " << i << ", positive \\epsilon" << endl;
			Tepsilon = &Ua;
			Bepsilon = &Da;
		}
		else
		{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: i = " << i << ", negative \\epsilon" << endl;
			Tepsilon = &invDa;
			Bepsilon = &invUa;
		}

		for (int j=0; j<2; j++) // \eta
		{
			matrix<int>* Teta;
			matrix<int>* Beta;
			if (j == 0)
			{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: j = " << j << ", positive \\eta" << endl;
				Teta = &Ua;
				Beta = &Da;
			}
			else
			{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: j = " << j << ", negative \\eta" << endl;
				Teta = &invDa;
				Beta = &invUa;
			}
			
			for (int k=0; k<2; k++) // \xi
			{
				matrix<int>* Txi;
				matrix<int>* Bxi;
				if (k == 0)
				{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: k = " << k << ", positive \\xi" << endl;
					Txi = &Ub;
					Bxi = &Db;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates: k = " << k << ", negative \\xi" << endl;
					Txi = &invDb;
					Bxi = &invUb;
				}
				
				bool R3_case_satisfied = R3_dominate_case(*Tepsilon, *Bepsilon, *Teta, *Beta, *Txi, *Bxi);
				
				if(i == j)
				{
					/* polarity of \epsilon and \eta crossings are the same, R3 case required to fail */
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  \\epsilon and \\eta crossings have the same polarity R3 case required to fail" << endl;

					if (R3_case_satisfied)
					{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  R3 case satisified, returning false" << endl;
						return false;
					}		
					else
					{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  R3 case fails, consistent with \\epsilon and \\eta crossings" << endl;
					}		

				}
				else
				{
					/* polarity of \epsilon and \eta crossings are different, R3 case required to be satisfied */
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  \\epsilon and \\eta crossings have different polarity R3 case required to be satisfied" << endl;

					if (!R3_case_satisfied)
					{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  R3 case not satisified, returning false" << endl;
						return false;
					}		
					else
					{
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  R3 case satisfied, consistent with \\epsilon and \\eta crossings" << endl;
					}		

				}				
			}
		}
	}
if (DEBUG >= DETAIL)
	debug << "singularly_dominates:  all R3 cases consistent with singular requirements, returning true" << endl;
	return true;
}


/* fully_dominates carries out a complete suite of tests corresponding to the generalized R3 move
   R_3(\epsilon, \eta; \xi), involving crossings of type C_i with polarity \epsilon and \eta 
   and a crossing of type C_j with polarity \xi.  If A is the switch corresponding to crossings of type 
   C_i and B is the switch corresponding to crossings of type C_j, the full suite of tests covers
   all eight relations of the form:
    
   (A^\epsilon x I) (I x B^\xi) (A^\eta x I) = (I x A^\eta) (B^\xi x I) (I x A^\epsilon)
    
   Each case is checked by evaluating the matrix representing the composition on
   the lhs and rhs.

   In the above relation, e.g. A^\epsilon is either A or invA depending on the polarity of \epsilon.
   
   The boolean argument "requirement" determines whether it is required that all the above relations are satisfied 
   (requirement == true) or whether it is required that none of the above relations are satisfied (requirement == false)
   
   The function returns true if the specified requirement is met and false otherwise.
*/
bool fully_dominates(matrix<int>& Ua, matrix<int>& Da, matrix<int>& Ub, matrix<int>& Db, bool requirement)
{

if (DEBUG >= DETAIL)
{
	debug << "fully_dominates: \\epsilon and \\eta switch" << endl;
	debug << "Ua = ";
	display (Ua, debug);
	debug << " Da = ";
	display (Da, debug);
	debug << endl;
	debug << "fully_dominates: \\xi switch" << endl;
	debug << "Ub = ";
	display (Ua, debug);
	debug << " Db = ";
	display (Da, debug);
	debug << endl;
	debug << "fully_dominates: requirement is " << (requirement? "true": "false") << "; that is, all domination cases are ";
	if (!requirement)
		debug << "not ";
	debug << "satisfied" << endl;
}

	/* calculate the inverse of A and B.  Given a switch S, invS(x,y) = (y_{\bar{x}},x^{\bar{y}}); 
	   we create matrices invSu and invSd so that the (x,y) entry of invSu is y^{\bar x} and 
	   the (x,y) entry of invSd is y_{\bar x}.

	   From the Reidemeister II move with a positive crossing followed by a negative crossing
	   we have that (b^a)^{\bar{a_b}}=b and (a_b)_{\bar{b^a}} = a, so that the (a_b, b^a) entry
	   of invSu is b and the (b^a, a_b) entry of invSd is a.
	*/
	matrix<int> invUa(n,n), invDa(n,n);	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int b_up_a = Ua[a][b];
		int a_down_b = Ua[b][a];
			
		invUa[a_down_b][b_up_a]=b;
		invDa[b_up_a][a_down_b]=a;		
	}

	matrix<int> invUb(n,n), invDb(n,n);	
	for (int a=0; a< n; a++)
	for (int b=0; b< n; b++)
	{
		int b_up_a = Ub[a][b];
		int a_down_b = Db[b][a];
			
		invUb[a_down_b][b_up_a]=b;
		invDb[b_up_a][a_down_b]=a;		
	}
				
	/* in the following loop clauses the value 0 will be taken to mean positive polarity
	   and the value 1 will be taken to mean negative polarity.
	   
	   For a switch S the operation at a crossing of positive polarity is S(x,y) = (y^x, x_y)
	   and at a crossing of negative polarity it is invS(x,y) = (y_{\bar x}, x^{\bar y}) so (working 
	   top down), the top operation is the up operation for a positive crossing and is the (inverse) down
	   operation for a negative crossing.  We therefore use Tepsilon and Bepsilon etc. to refer to the 
	   top and bottom operation matrices used at the epsilon crossing to cover both the positive and 
	   negative polarity cases.
	*/

	for (int i=0; i<2; i++) // \epsilon
	{
		matrix<int>* Tepsilon;
		matrix<int>* Bepsilon;
		if (i == 0)
		{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: i = " << i << ", positive \\epsilon" << endl;
			Tepsilon = &Ua;
			Bepsilon = &Da;
		}
		else
		{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: i = " << i << ", negative \\epsilon" << endl;
			Tepsilon = &invDa;
			Bepsilon = &invUa;
		}

		for (int j=0; j<2; j++) // \eta
		{
			matrix<int>* Teta;
			matrix<int>* Beta;
			if (j == 0)
			{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: j = " << j << ", positive \\eta" << endl;
				Teta = &Ua;
				Beta = &Da;
			}
			else
			{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: j = " << j << ", negative \\eta" << endl;
				Teta = &invDa;
				Beta = &invUa;
			}
			
			for (int k=0; k<2; k++) // \xi
			{
				matrix<int>* Txi;
				matrix<int>* Bxi;
				if (k == 0)
				{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: k = " << k << ", positive \\xi" << endl;
					Txi = &Ub;
					Bxi = &Db;
				}
				else
				{
if (DEBUG >= DETAIL)
	debug << "fully_dominates: k = " << k << ", negative \\xi" << endl;
					Txi = &invDb;
					Bxi = &invUb;
				}
				
				bool R3_case_satisfied = R3_dominate_case(*Tepsilon, *Bepsilon, *Teta, *Beta, *Txi, *Bxi);
				
				if (requirement == true && !R3_case_satisfied)
				{
if (DEBUG >= DETAIL)
	debug << "fully_dominates:  R3 case not satisified when requirement is true, returning false" << endl;
					return false;
				}
				else if (requirement == false && R3_case_satisfied)
				{
if (DEBUG >= DETAIL)
	debug << "fully_dominates:  R3 case satisified when requirement is false, returning false" << endl;
					return false;
				}		
				else
				{
if (DEBUG >= DETAIL)
	debug << "fully_dominates:  R3 case consistent with requirement" << endl;
				}		
			}
		}
	}
if (DEBUG >= DETAIL)
	debug << "fully_dominates:  all R3 cases consistent with requirement, returning true" << endl;
	return true;
}

void display_as_perms (matrix<int>& M, ostream& os, int format)
{
	/* we'll record which elements of the row remain to be written via row_flags */
	int row_flags[n];

	bool identity = true;
	/* first check to see if the matrix is the identity */
	for (int i=0; i< n && identity ; i++)
	{
		for (int j=0; j<n; j++)
			row_flags[j] = 1;

		bool found;
		int start;
	
		do
		{
			found = false;
			int j;
	
			/* look for another starting place in row_flags */
			for (j=0; j<n; j++)
			{
				if (row_flags[j])
				{
					start = j;
					row_flags[j] = 0;
					found = true;
					break;
				}
			}

			if (found)
			{
				if (M[i][j] != start)
				{
					identity = false;
					break;
				}
			}
		} while (found);
	}
	
	if (identity)
	{
		if (format == TeX_FORMAT)
			os << "\\iota";
		else
			os << "i";
	}
	else
	{
		/* we have to print out the matrix row by row, 
		   some of which may be the identity permutation
		*/
		os << "(";

		for (int i=0; i< n; i++)
		{
			for (int j=0; j<n; j++)
				row_flags[j] = 1;

			bool found;
			int start;
	
			identity = true;
			do
			{
				found = false;
				int j;
	
				/* look for another starting place in row_flags */
				for (j=0; j<n; j++)
				{
					if (row_flags[j])
					{
						start = j;
						row_flags[j] = 0;
						found = true;
						break;
					}
				}

				if (found)
				{
					if (M[i][j] != start)
					{
						identity = false;
						os << '(';
						os << j + (NUMBER_FROM_ZERO? 0:1);
			
						do
						{
							os << ' ';
							os << M[i][j] + (NUMBER_FROM_ZERO? 0:1);
								row_flags[M[i][j]] = 0;
								j = M[i][j];
						} while (M[i][j] != start);
					
						os << ')';
					}
				}
			} while (found);
		
			if (identity)
			{
				if (format == TeX_FORMAT)
					os << "\\iota";
				else
					os << "i";
			}
		
			if (i<n-1)
				os << " , ";
		}

		os << ")";
	}
}
