/************************************************************************

A quandle or rack search based on the birack search in nswitch.

 We define a switch structure comprised of a U D pair with entries numbered
 1,...,n rather than 0,...,n-1 so we can use a zero entry as an indication of 
 a value to be determined.  We speak of a prototype switch being a one containing 
 one or more zero entries.  We create a working list of prototype switches, 
 initialized to a single switch comprised entirely of zeros and a list of complete
 switches to hold the output of the search.
 
 Although for quandles and racks we don't need the D matrix, we'll leave the code 
 alone and only use the U matrix in this search.
 
 The search is a loop that continues until the prototype switch list is empty.  The
 loop takes the first prototype switch from the list and uses findzero and ratezero 
 to select an entry in U, then it uses avail to determine a set of values 
 available as a candidate for that entry and calls switchfill for each one to complete 
 as many other entries based on the candidate provided.
 
 The switchfill function is equivalent to Nelson's biqfill but for switches not 
 biquandles.  It returns a boolean to indicate whether it has found a contradiction
 to the quandle axioms with the specified value or not, and if not produces either a 
 fully complete switch stucture that can be added to the output list, or a prototype 
 switch structure that needs adding to the back of the working list.


A. Bartholomew May 2009


****************************************************************************/
using namespace std;

#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <list>
#include <ctype.h>
#include <stdio.h>
#include <vector>
#include <valarray>
#include <iomanip>

ofstream    debug;

#include <util.h>
#include <matrix.h>

#define OFF				0
#define BASIC			1
#define INTERMEDIATE	2
#define DETAIL 			3

#define UP_FUNCTION		1
#define DOWN_FUNCTION	2

unsigned int DEBUG = OFF;

struct Switch 
{

	Switch(matrix<int> u,matrix<int> d):U(u),D(d){}
	matrix<int> U;
	matrix<int> D;
};



int n; // size of finite field used in search
bool TeX_OUTPUT_DISPLAY = false;
bool OMIT_INVERSE_SYMMETRY = false;

void findzero(Switch& S, int& up_or_down, int& row, int& col);
int avail(Switch& S, int up_or_down, int row, int v);
bool switchfill (Switch& S);
bool completed_switch(Switch& S);
int ratezero (Switch& S, int up_or_down, int row, int col);
void display(matrix<int>& M, ostream& os, bool TeX, bool decremented=false);
bool valid_row_perm_value(matrix<int>& M, int row, int value);
bool check_list(Switch& filled_switch, list<Switch>& quandles);

int main(int argc, char* argv[])
{
	bool options;
	bool RACK_SEARCH = false; // the default is a quandle search
	bool NO_SYMMETRY_CHECK = false;
	bool OMIT_D_FROM_OUTPUT = false;
	bool NUMBER_OUTPUT_FROM_ZERO = false;
	
	string output_file = "nquandle.out";
	list<Switch> working_list;
	list<Switch> output_list;
	
	n=3; 
	
	if (argc >= 2)
	{
		if ( *(argv[1]) == '-')
			options = true;

		if (options)
		{
			char* dptr = strchr(argv[1], '#');
			if (dptr != 0)
			{
				DEBUG = BASIC;
				if (isdigit(*++dptr))
				{
					if (*dptr == '2')
						DEBUG = INTERMEDIATE;
					else if (*dptr == '3')
						DEBUG = DETAIL;
				}
			}

			if (strchr(argv[1], 'D'))
				OMIT_D_FROM_OUTPUT = true;

			if (strchr(argv[1], 'I'))
				OMIT_INVERSE_SYMMETRY = true;

			if (strchr(argv[1], 'n'))
			{
	    		char* c1 = strchr(argv[1],'n')+1;
	    		if (*c1 == '=')
	    		{
					char* c2 = ++c1 + 1;
					while(isdigit(*c2))
			    		c2++;
					get_number(n,c1);
	    		}
			}
			
			if (strchr(argv[1], 'r'))
				RACK_SEARCH = true;

			if (strchr(argv[1], 'T'))
				TeX_OUTPUT_DISPLAY = true;

			if (strchr(argv[1], 'X'))
				NO_SYMMETRY_CHECK = true;

			if (strchr(argv[1], 'Z'))
				NUMBER_OUTPUT_FROM_ZERO = true;
		}

	}
	else if (argc == 1 || argc > 3)
	{
		cout << "\n\nUsage: nquandle -n=<n>[#[1|2|3]DIrTXZ]" << endl;
		cout << "      n=<n> specify the size of the field" << endl;		
		cout << "      #[1|2|3] generate debug output" << endl;
		cout << "        optionally specify level: 1=BASIC, 2=INTERMEDIATE, 3=DETAIL (default BASIC)" << endl;
		cout << "      D omit the down operation (=Identity) from output" << endl;
		cout << "      I do not remove inverses in symmetry check" << endl;
		cout << "      r conduct a rack search rather than a quandle search" << endl;
		cout << "      T display matrix output and debug in TeX format" << endl;
		cout << "      X exclude summetry check" << endl;
		cout << "      Z number output matrices from zero" << endl;
		exit(0);
	}

if (DEBUG >= BASIC)
{
    /* establish a debug file */
    debug.open ("nquandle.dbg");

    if (!debug)
    {
        cout << "\nError opening debug file";
        exit(0);
    }
	else
	{
		debug << "debug from nquandle search for " << (RACK_SEARCH? "racks" : "quandles") << endl;
	}
}

	ofstream output(output_file.c_str());
	int counter = 1;

	if (!output)
	{
		cout << "\nError opening output file";
		exit(0);
	}
	
	/* The function switchfill needs a Switch structure to return switches */
	Switch filled_switch(matrix<int> (n,n), matrix<int> (n,n));
	
	/* Since the matrix constructor initializes to zero, we can use filled_switch 
	   to initialize the working list.  For a quandle search the diagonal of U
	   is always {1,...n}.
	*/
	if (!RACK_SEARCH)
	{
		for (int i=0; i< n; i++)
			filled_switch.U[i][i] = i+1;
	}
	
	working_list.push_back(filled_switch);
	
	/*main loop begins */
	while (working_list.size() != 0)
	{
		/* Find the best next candidate entry location in the first 
		   switch on the working list
		*/
if (DEBUG >= BASIC)
{
	debug << "main: prototype switch: U = ";
	display (working_list.begin()->U, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}

		int r,c;
		int up_or_down;
		findzero(*working_list.begin(),up_or_down,r,c);

if (DEBUG >= BASIC)		
{
	debug << "main: findzero returned " << (up_or_down == UP_FUNCTION? "U ": "D ") 
	      << "[" << r << "][" << c << "] " << endl;
}
	
	
		/* take each available value for this location and try to fill in as many other 
		  entries as possible.  The first candidate value in X={1,...,n} is 1 but the 
		  call to avail increments value so we step through all possible values in the
		  while clause.
		*/
		int value = 0; 
		
		while ((value = avail(*working_list.begin(),up_or_down,r,value+1)) != 0 )
		{
		
if (DEBUG >= BASIC)
{
	debug << "main: checking value " << value << " in " << (up_or_down == UP_FUNCTION? "U ": "D ") 
	      << "[" << r << "][" << c << "] " << endl;
}

			filled_switch = *working_list.begin();
			if (up_or_down == UP_FUNCTION)
				filled_switch.U[r][c] = value;				
			else
				filled_switch.D[r][c] = value;
							
			if (switchfill (filled_switch))
			{
				/* filled_switch contains the result of switchfill's work */
				if (completed_switch(filled_switch))
				{
				
if (DEBUG >= BASIC)
	debug << "main: switch completed, adding to output list" << endl;
	
					if (NO_SYMMETRY_CHECK)
						output_list.push_back(filled_switch);
					else
						check_list(filled_switch, output_list);
					
					cout << "#(" << working_list.size() << ")" << endl;
				}
				else
				{
				
if (DEBUG >= BASIC)
	debug << "main: switch not complete, adding to working list" << endl;
	
					working_list.push_back(filled_switch);
					cout << "\r" << working_list.size() << flush;
				}
			}
			else
			{
if (DEBUG >= BASIC)
	debug << "main: value " << value << " yields contradiction" << endl;
			}
		}
		
		/* We've now finished with the first item on the working_list */
		working_list.pop_front();			
	} // main loop end
	
	list<Switch>::iterator lptr = output_list.begin();
	while (lptr != output_list.end())
	{

		if (TeX_OUTPUT_DISPLAY)
		{
			output << "$$" << (RACK_SEARCH? "R" : "Q") << "^";
			output << n << "_{" << counter++ << "}\\quad ";
			output << "U=\\pmatrix{";
			for (int i=0; i< n; i++)
			{
				for (int j=0; j< n; j++)
					output << lptr->U[i][j] - (NUMBER_OUTPUT_FROM_ZERO? 1 : 0) << (j < n-1? " & ": "");
			
				if (i < n-1)
					output << " \\cr ";
			}

			if (!OMIT_D_FROM_OUTPUT)
			{
				/* write D as the identity function */
				output << "}\\quad D=\\pmatrix{";
				for (int i=0; i< n; i++)
				{
					for (int j=0; j< n; j++)
						output << j  + (NUMBER_OUTPUT_FROM_ZERO? 0 : 1) << (j < n-1? " & ": "");
				
					if (i < n-1)
						output << " \\cr ";
				}
			}

			output << "}\\quad ";	
			output << "$$" << endl;
		}
		else
		{
			output << "U = ";
			display(lptr->U,output,false, (NUMBER_OUTPUT_FROM_ZERO? true : false) );
			if (!OMIT_D_FROM_OUTPUT)
			{
				output << " D = ";
				for (int i=0; i< n; i++)
				{
					for (int j=0; j< n; j++)
						output << j  + (NUMBER_OUTPUT_FROM_ZERO? 0 : 1) << " ";

					if (i < n-1)
						output << " ";
				}
			}
			output << endl;
		}
		
		lptr++;
	}
	
	cout << endl;
	output.flush();
	output.close();

if (DEBUG >= BASIC)
{
    debug.flush();
    debug.close();
}

} /* end of main */

void findzero(Switch& S, int& up_or_down, int& row, int& col)
{
	/* initialize max_rating to -1 to allow for those early cases where
	   all zeros have zero rating, in which case we return the first
	   zero found
	*/
	int max_rating=-1; 
	
	row=col=0; //deals with the initial case when S is all zero

	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.U[i][j] == 0)
		{

if (DEBUG >= INTERMEDIATE)
	debug << "findzero: found zero in U[" << i << "][" << j << "] " << endl;
	
			int rating=	ratezero(S, UP_FUNCTION,i,j);
			
if (DEBUG >= INTERMEDIATE)
	debug << "findzero: rating = " << rating << endl;
			
			if (rating > max_rating)
			{
				max_rating = rating;
				row=i;
				col=j;
				up_or_down = UP_FUNCTION;
			}
		}
	}
}

bool completed_switch(Switch& S)
{
	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.U[i][j] == 0)
			return false;		
	}
	return true;
}

/* ratezero calculates the number of Y-B conditions that can be checked given 
   a value for the specified function entry.  It also checks whether the 
   corresponding row contains only two zeros, since setting one allows 
   the other to be checked by nature of the fact that rows are permutations
*/
int ratezero(Switch& S, int up_or_down, int row, int col)
{
	int rating=0;

if (DEBUG >= DETAIL)
	debug << "ratezero: rating " << (up_or_down == UP_FUNCTION? "U ": "D ") << "[" << row << "][" << col << "] " << endl;
	
	/* This comments to this code have to be read with the understanding that X={1,...,n} 
	   not {0,...n-1}, however, we are looking for non-zero entries in S.U and S.D that allow
	   us to check more of the Y-B condition.  The comments are written as though X={0,...n-1}
	   as it makes them compatible with theory.
	*/

	if (up_or_down == UP_FUNCTION)
	{
	
	
		int a=row;
		int b=col;
		int row_count=0;
		
		/* The rack YB relation is c^{ba} = (c^a)^{b^a} for all a,b,c.
		   That is U[a][U[b][c]] = U[U[a][b]][U[a][c]]
		   
		   Therefore, for a given value of row a and column b avail will be providing
		   various values of b^a, so for each c, increment the rating if we already 
		   have c^b and c^a since they will allow us to check the rack YB relation.
		*/
		for (int c=0; c<n; c++)
		{
			if (S.U[b][c] && S.U[a][c])
			{
if (DEBUG >= DETAIL)
	debug << "ratezero: up interchange check possible with c = " << c << endl;
				rating++;
			}
				
			if(S.U[row][c]==0)
				row_count++;
		}
		
		if(row_count==2)
		{
if (DEBUG >= DETAIL)
	debug << "ratezero: row_count == 2" << endl;
			rating++;
		}
	}
	else 
	{
		cout << "\nError! up_or_down != UP_FUNCTION in ratezero!" << endl;
		exit(0);	
	}
	
	return rating;
}

/* avail determines the smallest valid value for the specified row that is >= v, 
   or zero if none is found.  A value is valid if it does not contradict the
   requirement for rows of U and D to be permutations. (This is different to Nelson's
   corresponding function in that he is looking for biquandles so has an additonal
   column condition on validity.)
*/
int avail(Switch& S, int up_or_down, int row, int v)
{
	matrix<int>& M = (up_or_down == UP_FUNCTION ? S.U : S.D);

if (DEBUG >= BASIC)
	debug << "avail: checking row " << row << " of " << (up_or_down == UP_FUNCTION? "U ": "D ") << "from " << v << endl;
	
	for (int c=v; c<=n; c++) // c is a candidate value
	{
	
if (DEBUG >= BASIC)
	debug << "avail: candidate value " << c;
	
		bool found = false;
		for (int i=0; i<n; i++)
		{
			if (M[row][i] == c)
			{
				found = true;
if (DEBUG >= BASIC)
	debug << " found in column " << i << endl;
				break;			
			}
		}

		if (!found)
		{
if (DEBUG >= BASIC)
	debug << " not found, returning " << c << endl;
				return c;
		}			
	}
	
if (DEBUG >= BASIC)
	debug << "avail: no candidate found, returning zero" << endl;
	
	return 0;
}

bool switchfill (Switch& S)
{
	matrix<int>& U = S.U;
	
	bool no_contradiction = true;
	bool check_again = true;
	
	/* Check whether any zeros may be filled in from the rack YB condition */
	while (check_again)
	{
		check_again = false;
		
if (DEBUG >= INTERMEDIATE)
{
	debug << "switchfill: starting check of triples a b c" << endl;
	debug << "switchfill: prototype switch: U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	
		for (int a=0; a < n; a++)
		for (int b=0; b < n; b++)
		for (int c=0; c < n; c++)
		{
		
if (DEBUG >= INTERMEDIATE)
	debug << "switchfill: a = " << a << " b = " << b << " c = " << c << endl;
	
			/* the YB condition requires U[a][U[b][c]] = U[U[a][b]][U[a][c]] */
			if (U[b][c] && U[a][b] && U[a][c])
			{
				int lhs = U[a][U[b][c]-1];
				int rhs = U[U[a][b]-1][U[a][c]-1];

if (DEBUG >= DETAIL)
{
	debug << "switchfill:   YB lhs: U[" << b << "][" << c << "]-1 = " << U[b][c]-1 << endl;
	debug << "switchfill:   YB rhs: U[" << a << "][" << b << "]-1 = " << U[a][b]-1
          << " U[" << a << "][" << c << "]-1 = " << U[a][c]-1 << endl;
}

if (DEBUG >= INTERMEDIATE)
	debug << "switchfill:   YB check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(U,U[a][b]-1,lhs))
					{
						U[U[a][b]-1][U[a][c]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "switchfill:     setting U[" << U[a][b]-1 << "][" << U[a][c]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << U[a][b]-1 << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(U,a,rhs))
					{
						U[a][U[b][c]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "switchfill:     setting U[" << a << "][" << U[b][c]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << a << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}								
			}
		}
		
		/* Check whether any rows have a single non-zero entry 
		   and fill in based on the permutation requirement */
		for (int i=0; i<n; i++)
		{
			int U_row_count = 0;
			int Ui, Uj;
			
			for (int j=0; j < n && U_row_count < 2; j++)
			{
				if (U[i][j] == 0)
				{
					Ui=i;
					Uj=j;
					U_row_count++;
				}
			}
			
			if (U_row_count == 1)
			{
if (DEBUG >= BASIC)
	debug << "switchfill:   row " << i << " of U contains a single zero" << endl;					
				int value = avail(S,UP_FUNCTION,i,1);
				if (value != 0)
				{
					U[Ui][Uj] = value;
					check_again = true;
if (DEBUG >= BASIC)
	debug << "switchfill:   set U[" << Ui << "][" << Uj << "] = " << value << endl;					
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "switchfill:   no available value to complete row " << i << " in U" << endl;
					goto contradiction_bailout;
				}
					
			}			
		}				
	}	
	contradiction_bailout:	

if (DEBUG >= INTERMEDIATE)
{
	debug << "switchfill: returning (prototype) switch U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	return no_contradiction;
}

void display(matrix<int>& M, ostream& os, bool TeX, bool decremented)
{
	if (TeX)
		os << "$$\\pmatrix{";
		
	for (int i=0; i< n; i++)
	{
		for (int j=0; j< n; j++)
			os << (decremented? M[i][j]-1: M[i][j]) << ((TeX && j < n-1)? " & ": " ");
			
		if (i < n-1)
			os << (TeX? " \\cr ": " ");
	}

	if (TeX)
		os << "}$$";
}

/* valid_row_perm_value returns true if value does not appear in the 
   specified row of M and false otherwise.
*/
bool valid_row_perm_value(matrix<int>& M, int row, int value)
{
	for (int i=0; i< n; i++)
	{
		if (M[row][i] == value)
			return false;
	}
	return true;
}

/* check_list looks for permutation and Obv symmetry in the list.  It returns true if the up action in filled_switch 
   or one of its symmetries does not appear in the specified list.

   We use the term Sigma(S) rather than sigma_sharp(S) and calculate Sigma(S) as in the up action isomorphic
   switches.
*/
bool check_list( Switch& filled_switch, list<Switch>& quandles)
{
	list<Switch>::iterator lptr=quandles.begin();
	matrix <int>& U = filled_switch.U;
	matrix <int> SigmaU(n,n);
	matrix <int> ObvU(n,n);
				
	/* the identity permutation will check U as provided */
	vector<int> perm(n), inv_perm(n);
	for (int i=0; i< n; i++)
		perm[i]=i;
		
	do
	{
						
		for (int i=0; i< n; i++)
			inv_perm[perm[i]]=i;

if (DEBUG >= INTERMEDIATE)
{
	debug << "    perm symmetry check with sigma = ";
	for (int i=0; i< n; i++)
		debug << perm[i] << ' ';
	debug << ", inv_sigma = ";
	for (int i=0; i< n; i++)
		debug << inv_perm[i] << ' ';
	debug << endl;
}
							
		/* create Sigma(S) */
		for (int a=0; a< n; a++)
		for (int b=0; b< n; b++)
		{
			int inv_sigma_a__up__inv_sigma_b = U[inv_perm[b]][inv_perm[a]]-1;
			SigmaU[b][a] = perm[inv_sigma_a__up__inv_sigma_b]+1;
		}
					
if (DEBUG >= INTERMEDIATE)
{
	debug << "    SigmaU = ";
	display (SigmaU, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;
}

		/* calculate Obv(Sigma), we have to create a matrix ObvU so that the (x,y) entry of ObvU is y^{\bar x}.
		   
		   From the Reidemeister II move with a positive crossing followed by a negative crossing we have that 
		   (b^a)^{\bar a}=b, so that the (a, b^a) entry of ObvU is b.
		*/
					
		for (int a=0; a< n; a++)
		for (int b=0; b< n; b++)
		{
			int b_up_a = SigmaU[a][b]-1;					
			ObvU[a][b_up_a]=b+1;
		}

if (DEBUG >= INTERMEDIATE)
{
	debug << "    Obv(Sigma(S)):" << endl;
	debug << "    ObvU = ";
	display (ObvU, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;
}

		lptr=quandles.begin();
		while (lptr != quandles.end())
		{
			if (lptr-> U == SigmaU)
			{

if (DEBUG >= BASIC)
{
	debug << "    equality (i.e. Sigma(S)) found with sigma = ";
	for (int i=0; i< n; i++)
		debug << perm[i] << ' ';
	debug << " in list" << endl;
}	
				return false;
			}			
							
			if (!OMIT_INVERSE_SYMMETRY && lptr-> U == ObvU)
			{
		
if (DEBUG >= BASIC)
	debug << "    Obv found in list" << endl;

				return false;	
			}

			lptr++;
		}
	} while (next_permutation(perm.begin(),perm.end()));


	/* We've got a genuinely new biquandle, add it to the list */
	quandles.push_back(filled_switch);
	
if (DEBUG >= BASIC)
	debug << "  distinct symmetry class representative, adding to list" << endl;

	return true;	
}
