/************************************************************************

A birack search based on the biquandle search algorithm of Sam Nelson and
John Vo in the paper Matrices and Finite Biquandles (arXiv:math/0601145v4 
April 2006).


 We define a switch structure comprised of a U D pair with entries numbered
 1,...,n rather than 0,...,n-1 so we can use a zero entry as an indication of 
 a value to be determined.  We speak of a prototype switch being a one containing 
 one or more zero entries.  We create a working list of prototype switches, 
 initialized to a single switch comprised entirely of zeros and a list of complete
 switches to hold the output of the search.
 
 The search is a loop that continues until the prototype switch list is empty.  The
 loop takes the first prototype switch from the list and uses findzero and ratezero 
 to select an entry in either U or D, then it uses avail to determine a set of values 
 available as a candidate for that entry and calls switchfill for each one to complete 
 as many other entries based on the candidate provided.
 
 The switchfill function is equivalent to Nelson's biqfill but for switches not 
 biquandles.  It returns a boolean to indicate whether it has found a contradiction
 to the switch axioms with the specified value or not, and if not produces either a 
 fully complete switch stucture that can be added to the output list, or a prototype 
 switch structure that needs adding to the back of the working list.

This started out as an attempt to create a search for switches, hence the name,
but we needed some conditions to limt the number of combinations considered so 
added the condition that rows of U and D were permutations, hence it became a 
search for biracks rather than switches.

A. Bartholomew April 2009

Around 2013 the code was amended to search for biracks using what was referred to as the
SIDEWAYS_SEARCH, what is now referred to as F_YANG_BAXTER_SEARCH.  In 2014, the code was updated
to bring it in line with the main biquandle.cpp module to use the terms prefixed by S_ or F_
according to the understanding gained with the generalized biquandle theory.
****************************************************************************/
using namespace std;

#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <list>
#include <ctype.h>
#include <stdio.h>
#include <vector>
#include <valarray>
#include <iomanip>

ofstream    debug;

#include <util.h>
#include <matrix.h>

#define OFF				0
#define BASIC			1
#define INTERMEDIATE	2
#define DETAIL 			3

#define UP_FUNCTION		1
#define DOWN_FUNCTION	2

unsigned int DEBUG = OFF;

struct Switch 
{

	Switch(matrix<int> u,matrix<int> d):U(u),D(d){}
	matrix<int> U;
	matrix<int> D;
};



int n; // size of finite field used in search
bool TeX_OUTPUT_DISPLAY = false;
bool NUMBER_FROM_ZERO = false;
bool S_YANG_BAXTER_SEARCH = false;

void findzero(Switch& S, int& up_or_down, int& row, int& col);
int avail(Switch& S, int up_or_down, int row, int v);
bool switchfill (Switch& S);
bool completed_switch(Switch& S);
int ratezero (Switch& S, int up_or_down, int row, int col);
void display(matrix<int>& M, ostream& os, bool TeX);
bool valid_row_perm_value(matrix<int>& M, int row, int value);

int main(int argc, char* argv[])
{
	bool options;
	bool READ_SEED_PATTERNS = false;
	
	string output_file = "nswitch.out";
	list<Switch> working_list;
	list<Switch> output_list;
	
	n=3; 
	
	if (argc >= 2)
	{
		if ( *(argv[1]) == '-')
			options = true;

		if (options)
		{
			char* dptr = strchr(argv[1], '#');
			if (dptr != 0)
			{
				DEBUG = BASIC;
				if (isdigit(*++dptr))
				{
					if (*dptr == '2')
						DEBUG = INTERMEDIATE;
					else if (*dptr == '3')
						DEBUG = DETAIL;
				}
			}

			if (strchr(argv[1], 'n'))
			{
	    		char* c1 = strchr(argv[1],'n')+1;
	    		if (*c1 == '=')
	    		{
					char* c2 = ++c1 + 1;
					while(isdigit(*c2))
			    		c2++;
					get_number(n,c1);
	    		}
			}
			
			if (strchr(argv[1], 'R'))
				READ_SEED_PATTERNS = true;

			if (strchr(argv[1], 'S'))
				S_YANG_BAXTER_SEARCH = true;
			
			if (strchr(argv[1], 'T'))
				TeX_OUTPUT_DISPLAY = true;

			if (strchr(argv[1], 'Z'))
				NUMBER_FROM_ZERO = true;
		}

	}
	else if (argc == 1 || argc > 3)
	{
		cout << "\n\nUsage: nswitch -n=<n>[#[1|2|3]RSTZ <input_file>]" << endl;
		cout << "      n=<n> specify the size of the field (default 3)" << endl;		
		cout << "      #[1|2|3] generate debug output" << endl;
		cout << "        optionally specify level: 1=BASIC, 2=INTERMEDIATE, 3=DETAIL (default BASIC)" << endl;
		cout << "      R read seed patterns for U from <input_file>" << endl;
		cout << "      S use the S-Yang-Baxter relations rather than the F-Yang-Baxter relations" << endl;
		cout << "      T display matrix output in TeX format" << endl;
		cout << "      Z input and output X_n={0,...,n-1} rather than X_n={1,...,n}" << endl;
		exit(0);
	}

if (DEBUG >= BASIC)
{
    /* establish a debug file */
    debug.open ("nswitch.dbg");

    if (!debug)
    {
        cout << "\nError opening debug file";
        exit(0);
    }
}

	ofstream output(output_file.c_str());
	int counter = 1;

	if (!output)
	{
		cout << "\nError opening output file";
		exit(0);
	}
	
	/* The function switchfill needs a Switch structure to return switches */
	Switch filled_switch(matrix<int> (n,n), matrix<int> (n,n));
	
	if (READ_SEED_PATTERNS)
	{
		ifstream input(argv[2]);	
		if (input)
		{
			string next_line;
			while (getline(input,next_line))
			{
				istringstream iss(next_line);
				char ch;
				do
				{
					iss >> ch;	
				} while (!(ch == 'U'));
								
				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					do
					{
						iss >> ch;	
						if (isdigit(ch))
							iss.putback(ch);
					} while (!isdigit(ch));
					iss >> filled_switch.U[i][j];

					/* The code is written with X_n={1,...,n} to use 0 as a blank
					   if the input is not in that form adjust it */
					if (NUMBER_FROM_ZERO)
						filled_switch.U[i][j]++;
				}

				for (int i=0; i< n; i++)
				for (int j=0; j< n; j++)
				{
					filled_switch.D[i][j] = 0;
				}

if (DEBUG >= BASIC)
{
	debug << "Read seed up action from input file: U = ";
	display (filled_switch.U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (filled_switch.D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;
}
				working_list.push_back(filled_switch);		
			}
		}
		else
		{
			cout << "\nError opening input file" << endl;
			exit(0);
		}		
	}
	else
	{	
		/* Since the matrix constructor initializes to zero, we can use filled_switch 
		   to initialize the working list */
		working_list.push_back(filled_switch);
	}
	
	/*main loop begins */
	while (working_list.size() != 0)
	{
		/* Find the best next candidate entry location in the first 
		   switch on the working list
		*/
if (DEBUG >= BASIC)
{
	debug << "main: prototype switch: U = ";
	display (working_list.begin()->U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (working_list.begin()->D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}

		int r,c;
		int up_or_down;
		findzero(*working_list.begin(),up_or_down,r,c);

if (DEBUG >= BASIC)		
{
	debug << "main: findzero returned " << (up_or_down == UP_FUNCTION? "U ": "D ") 
	      << "[" << r << "][" << c << "] " << endl;
}
	
	
		/* take each available value for this location and try to fill in as many other 
		  entries as possible.  The first candidate value in X={1,...,n} is 1 but the 
		  call to avail increments value so we step through all possible values in the
		  while clause.
		*/
		int value = 0; 
		
		while ((value = avail(*working_list.begin(),up_or_down,r,value+1)) != 0 )
		{
		
if (DEBUG >= BASIC)
{
	debug << "main: checking value " << value << " in " << (up_or_down == UP_FUNCTION? "U ": "D ") 
	      << "[" << r << "][" << c << "] " << endl;
}

			filled_switch = *working_list.begin();
			if (up_or_down == UP_FUNCTION)
				filled_switch.U[r][c] = value;				
			else
				filled_switch.D[r][c] = value;
							
			if (switchfill (filled_switch))
			{
				/* filled_switch contains the result of switchfill's work */
				if (completed_switch(filled_switch))
				{
				
if (DEBUG >= BASIC)
	debug << "main: switch completed, adding to output list" << endl;
	
					output_list.push_back(filled_switch);
					
					cout << "#(" << working_list.size() << ")" << endl;
				}
				else
				{
				
if (DEBUG >= BASIC)
	debug << "main: switch not complete, adding to working list" << endl;
	
					working_list.push_back(filled_switch);
//					cout << "." << flush;
					cout << "\r" << working_list.size() << flush;
				}
			}
			else
			{
if (DEBUG >= BASIC)
	debug << "main: value " << value << " yields contradiction" << endl;
			}
		}
		
		/* We've now finished with the first item on the working_list */
		working_list.pop_front();			
	} // main loop end
		
	list<Switch>::iterator lptr = output_list.begin();
	while (lptr != output_list.end())
	{

		if (TeX_OUTPUT_DISPLAY)
		{
			output << "$$S";
			output << n << "_{" << counter++ << "}\\quad ";
			output << "U=\\pmatrix{";
			for (int i=0; i< n; i++)
			{
				for (int j=0; j< n; j++)
					output << lptr->U[i][j] + (NUMBER_FROM_ZERO? -1: 0) << (j < n-1? " & ": "");
			
				if (i < n-1)
					output << " \\cr ";
			}
		
			output << "}\\quad D=\\pmatrix{";
			for (int i=0; i< n; i++)
			{
				for (int j=0; j< n; j++)
					output << lptr->D[i][j] + (NUMBER_FROM_ZERO? -1: 0) << (j < n-1? " & ": "");
				
				if (i < n-1)
					output << " \\cr ";
			}

			output << "}\\quad ";	
			output << "$$" << endl;
		}
		else
		{
			output << "U = ";
			display(lptr->U,output,false);
			output << " D = ";
			output << " ";
			display(lptr->D,output,false);
			output << endl;
		}
		
		lptr++;
	}
	
	cout << endl;
	output.flush();
	output.close();

if (DEBUG >= BASIC)
{
    debug.flush();
    debug.close();
}

} /* end of main */

void findzero(Switch& S, int& up_or_down, int& row, int& col)
{
	/* initialize max_rating to -1 to allow for those early cases where
	   all zeros have zero rating, in which case we return the first
	   zero found
	*/
	int max_rating=-1; 
	
	row=col=0; //deals with the initial case when S is all zero

	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.U[i][j] == 0)
		{

if (DEBUG >= INTERMEDIATE)
	debug << "findzero: found zero in U[" << i << "][" << j << "] " << endl;
	
			int rating=	ratezero(S, UP_FUNCTION,i,j);
			
if (DEBUG >= INTERMEDIATE)
	debug << "findzero: rating = " << rating << endl;
			
			if (rating > max_rating)
			{
				max_rating = rating;
				row=i;
				col=j;
				up_or_down = UP_FUNCTION;
			}
		}
	}

	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.D[i][j] == 0)
		{

if (DEBUG >= INTERMEDIATE)
	debug << "findzero: found zero in D[" << i << "][" << j << "] " << endl;
	
	
			int rating = ratezero(S, DOWN_FUNCTION,i,j);
			
if (DEBUG >= INTERMEDIATE)
	debug << "findzero rating = " << rating << endl;
	
			if (rating > max_rating)
			{
				max_rating = rating;
				row=i;
				col=j;
				up_or_down = DOWN_FUNCTION;
			}
		}
	}
}

bool completed_switch(Switch& S)
{
	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.U[i][j] == 0)
			return false;		
	}
	
	for (int i=0; i<n; i++)
	for (int j=0; j<n; j++)
	{
		if (S.D[i][j] == 0)
			return false;		
	}	
	return true;
}

/* S_ratezero calculates the number of Y-B conditions that can be checked given 
   a value for the specified function entry.  It also checks whether the 
   corresponding row contains only two zeros, since setting one allows 
   the other to be checked by nature of the fact that rows are permutations
*/
int S_ratezero(Switch& S, int up_or_down, int row, int col)
{
	int rating=0;

if (DEBUG >= DETAIL)
	debug << "S_ratezero: rating " << (up_or_down == UP_FUNCTION? "U ": "D ") << "[" << row << "][" << col << "] " << endl;
	
	/* This comments to this code have to be read with the understanding that X={1,...,n} 
	   not {0,...n-1}, however, we are looking for non-zero entries in S.U and S.D that allow
	   us to check more of the Y-B condition.  The comments are written as though X={0,...n-1}
	   as it makes them compatible with theory.
	*/

	if (up_or_down == UP_FUNCTION)
	{
	
		int a=col;
		int b=row;
		int row_count=0;
		
		/* For a given value of a^b, for each c, increment the rating if
		   we already have b^c, c_b and a^{c_b}, since these allow us to check the 
		   up interchange.
		*/
		for (int c=0; c<n; c++)
		{
			if (S.U[c][b] && S.D[b][c] && S.U[S.D[b][c]-1][a])
			{
if (DEBUG >= DETAIL)
	debug << "S_ratezero: up interchange check possible" << endl;
				rating++;
			}
				
			if(S.U[row][c]==0)
				row_count++;
		}
		
		if(row_count==2)
		{
if (DEBUG >= DETAIL)
	debug << "S_ratezero: row_count == 2" << endl;
			rating++;
		}
	}
	else 
	{
		int a=col;
		int b=row;
		int row_count=0;
		
		/* For a given value of a_b, for each c, increment the rating if
		   we already have c^b, b_c and a_{c^b}, since these allow us to check the 
		   down interchange.  Also increment the rating if we have b^a, a^c
		   c_a, c_{b^a} and b^{c_a}, since these allow us to check the rule of five.
		*/
		for (int c=0; c<n; c++)
		{
			if (S.D[c][b] && S.U[b][c] && S.D[S.U[b][c]-1][a])
			{
if (DEBUG >= DETAIL)
	debug << "S_ratezero: down interchange check possible" << endl;
				rating++;
			}
				
			if (S.U[a][b] && S.U[c][a] && S.D[a][c] && S.D[S.U[a][b]-1][c] && S.U[S.D[a][c]-1][b] ) 
			{
if (DEBUG >= DETAIL)
	debug << "S_ratezero: rule of five check possible" << endl;			
			rating++;
			}
			
			if (S.D[row][c] == 0)
				row_count++;
		}
		
		if (row_count == 2)
		{
if (DEBUG >= DETAIL)
	debug << "S_ratezero: row_count == 2" << endl;
			rating++;
		}
	}
	
	return rating;
}

/* F_ratezero calculates the number of Y-B conditions that can be checked given 
   a value for the specified function entry.  It also checks whether the 
   corresponding row contains only two zeros, since setting one allows 
   the other to be checked by nature of the fact that rows are permutations
*/
int F_ratezero(Switch& S, int up_or_down, int row, int col)
{
	int rating=0;

if (DEBUG >= DETAIL)
	debug << "F_ratezero: rating " << (up_or_down == UP_FUNCTION? "U ": "D ") << "[" << row << "][" << col << "] " << endl;
	
	/* This comments to this code have to be read with the understanding that X={1,...,n} 
	   not {0,...n-1}, however, we are looking for non-zero entries in S.U and S.D that allow
	   us to check more of the Y-B condition.  The comments are written as though X={0,...n-1}
	   as it makes them compatible with theory.
	*/

	if (up_or_down == UP_FUNCTION)
	{
	
	
		int a=col;
		int b=row;
		int row_count=0;
		
		/* For a given value of a^b, for each c, increment the rating if
		   we already have c_b, b^c and a^c , since these allow us to check the 
		   up interchange.
		*/
		for (int c=0; c<n; c++)
		{

			if (S.D[b][c] && S.U[c][b] && S.U[c][a])
			{
if (DEBUG >= DETAIL)
	debug << "F_ratezero: up interchange check possible" << endl;
				rating++;
			}
				
			if(S.U[row][c]==0)
				row_count++;
		}
		
		if(row_count==2)
		{
if (DEBUG >= DETAIL)
	debug << "F_ratezero: row_count == 2" << endl;
			rating++;
		}
	}
	else 
	{
		int a=col;
		int b=row;
		int row_count=0;
		
		/* For a given value of a_b, for each c, increment the rating if
		   we already have c^b, b_c and a_c, since these allow us to check the 
		   down interchange.  Also increment the rating if we have c_b, b^c
		   and a^c, since these allow us to check the rule of four.
		*/
		for (int c=0; c<n; c++)
		{
			if (S.U[b][c] && S.D[c][b] && S.D[c][a])
			{
if (DEBUG >= DETAIL)
	debug << "F_ratezero: down interchange check possible" << endl;
				rating++;
			}
				
			if (S.D[b][c] && S.U[c][b] && S.U[c][a])
			{
if (DEBUG >= DETAIL)
	debug << "F_ratezero: rule of four check possible" << endl;			
			rating++;
			}
			
			if (S.D[row][c] == 0)
				row_count++;
		}
		
		if (row_count == 2)
		{
if (DEBUG >= DETAIL)
	debug << "F_ratezero: row_count == 2" << endl;
			rating++;
		}
	}
	
	return rating;
}

int ratezero(Switch& S, int up_or_down, int row, int col)
{
	if (S_YANG_BAXTER_SEARCH)
		return S_ratezero(S,up_or_down,row,col);	
	else
		return F_ratezero(S,up_or_down,row,col);
}

/* avail determines the smallest valid value for the specified row that is >= v, 
   or zero if none is found.  A value is valid if it does not contradict the
   requirement for rows of U and D to be permutations. (This is different to Nelson's
   corresponding function in that he is looking for biquandles so has an additonal
   column condition on validity.)
*/
int avail(Switch& S, int up_or_down, int row, int v)
{
	matrix<int>& M = (up_or_down == UP_FUNCTION ? S.U : S.D);

if (DEBUG >= BASIC)
	debug << "avail: checking row " << row << " of " << (up_or_down == UP_FUNCTION? "U ": "D ") << "from " << v << endl;
	
	for (int c=v; c<=n; c++) // c is a candidate value
	{
	
if (DEBUG >= BASIC)
	debug << "avail: candidate value " << c;
	
		bool found = false;
		for (int i=0; i<n; i++)
		{
			if (M[row][i] == c)
			{
				found = true;
if (DEBUG >= BASIC)
	debug << " found in column " << i << endl;
				break;			
			}
		}

		if (!found)
		{
if (DEBUG >= BASIC)
	debug << " not found, returning " << c << endl;
				return c;
		}			
	}
	
if (DEBUG >= BASIC)
	debug << "avail: no candidate found, returning zero" << endl;
	
	return 0;
}

bool S_switchfill (Switch& S)
{
	matrix<int>& U = S.U;
	matrix<int>& D = S.D;
	
	bool no_contradiction = true;
	bool check_again = true;
	
	/* Check whether any zeros may be filled in from the up down and rule of five conditions */
	while (check_again)
	{
		check_again = false;
		
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill: starting check of triples a b c" << endl;
	debug << "S_switchfill: prototype switch: U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	
		for (int a=0; a < n; a++)
		for (int b=0; b < n; b++)
		for (int c=0; c < n; c++)
		{
		
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill: a = " << a << " b = " << b << " c = " << c << endl;
	
			/* the up interchange requires U[c][U[b][a]] == U[U[c][b]][U[D[b][c]][a]] */
			if (U[b][a] && U[c][b] && D[b][c] && U[D[b][c]-1][a])
			{
				int lhs = U[c][U[b][a]-1];
				int rhs = U[U[c][b]-1][U[D[b][c]-1][a]-1];

if (DEBUG >= DETAIL)
{
	debug << "S_switchfill:   up interchange lhs: U[" << b << "][" << a << "]-1 = " << U[b][a]-1 << endl;
	debug << "S_switchfill:   up interchange rhs: U[" << c << "][" << b << "]-1 = " << U[c][b]-1
          << " D[" << b << "][" << c << "]-1 = " << D[b][c]-1
          << " U[D[" << b << "][" << c << "]-1][" << a << "]-1 = " << U[D[b][c]-1][a]-1 << endl;
}

if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:   up interchange check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(U,U[c][b]-1,lhs))
					{
						U[U[c][b]-1][U[D[b][c]-1][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting U[" << U[c][b]-1 << "][" << U[D[b][c]-1][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << U[c][b]-1 << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(U,c,rhs))
					{
						U[c][U[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting U[" << c << "][" << U[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << c << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}								
			}
		
			/* the down interchange requires D[c][D[b][a]] == D[D[c][b]][D[U[b][c]][a]] */
			if (D[b][a] && D[c][b] && U[b][c] && D[U[b][c]-1][a])
			{
				int lhs = D[c][D[b][a]-1];
				int rhs = D[D[c][b]-1][D[U[b][c]-1][a]-1];
				
if (DEBUG >= DETAIL)
{
	debug << "S_switchfill:   down interchange lhs: D[" << b << "][" << a << "]-1 = " << D[b][a]-1 << endl;
	debug << "S_switchfill:   down interchange rhs: D[" << c << "][" << b << "]-1 = " << D[c][b]-1
          << " U[" << b << "][" << c << "]-1 = " << U[b][c]-1
          << " D[U[" << b << "][" << c << "]-1][" << a << "]-1 = " << D[U[b][c]-1][a]-1 << endl;
}

if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:   down interchange check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(D,D[c][b]-1,lhs))
					{
						D[D[c][b]-1][D[U[b][c]-1][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting D[" << D[c][b]-1 << "][" << D[U[b][c]-1][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << D[c][b]-1 << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(D,c,rhs))
					{
						D[c][D[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting D[" << c << "][" << D[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << c << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}						
			}
			
			/* the rule of five requires that U[D[U[a][b]][c]][D[b][a]] == D[U[D[a][c]][b]][U[c][a]] */
			if (U[a][b] && D[U[a][b]-1][c] && D[a][c] && U[D[a][c]-1][b] && D[b][a] && U[c][a])
			{
				int lhs = U[D[U[a][b]-1][c]-1][D[b][a]-1];
				int rhs = D[U[D[a][c]-1][b]-1][U[c][a]-1];
				
if (DEBUG >= DETAIL)
{
	debug << "S_switchfill:   rule of five lhs: U[" << a << "][" << b << "] = " << U[a][b]
	      << " D[U[" << a << "][" << b << "]-1][" << c << "]-1 = " << D[U[a][b]-1][c]-1
          << " D[" << b << "][" << a << "] = " << D[b][a]	<< endl;
	debug << "S_switchfill:   rule of five rhs: D[" << a << "][" << c << "] = " << D[a][c] 
	      << " U[D[" << a << "][" << c << "]-1][" << b << "]-1 = " << U[D[a][c]-1][b]-1
          << " U[" << c << "][" << a << "] = " << U[c][a]	<< endl;
}
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:   rule of five check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(D,U[D[a][c]-1][b]-1,lhs))
					{
						D[U[D[a][c]-1][b]-1][U[c][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting D[" << U[D[a][c]-1][b]-1 << "][" << U[c][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << U[D[a][c]-1][b]-1 << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(U,D[U[a][b]-1][c]-1,rhs))
					{
						U[D[U[a][b]-1][c]-1][D[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "S_switchfill:     setting U[" << D[U[a][b]-1][c]-1 << "][" << D[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << D[U[a][b]-1][c]-1 << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}						
			}
						
		}
		
		/* Check whether any rows have a single non-zero entry 
		   and fill in based on the permutation requirement */
		for (int i=0; i<n; i++)
		{
			int U_row_count = 0;
			int D_row_count = 0;
			int Ui, Uj, Di, Dj;
			
			for (int j=0; j < n && U_row_count < 2; j++)
			{
				if (U[i][j] == 0)
				{
					Ui=i;
					Uj=j;
					U_row_count++;
				}
			}
			
			if (U_row_count == 1)
			{
if (DEBUG >= BASIC)
	debug << "S_switchfill:   row " << i << " of U contains a single zero" << endl;					
				int value = avail(S,UP_FUNCTION,i,1);
				if (value != 0)
				{
					U[Ui][Uj] = value;
					check_again = true;
if (DEBUG >= BASIC)
	debug << "S_switchfill:   set U[" << Ui << "][" << Uj << "] = " << value << endl;					
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "S_switchfill:   no available value to complete row " << i << " in U" << endl;
					goto contradiction_bailout;
				}
					
			}
			
			for (int j=0; j < n && D_row_count < 2; j++)
			{
				if(D[i][j] == 0)
				{
					Di=i;
					Dj=j;
					D_row_count++;
				}
			}
			
			if (D_row_count == 1)
			{

if (DEBUG >= BASIC)
	debug << "S_switchfill:   row " << i << " of D contains a single zero" << endl;					

				int value = avail(S,DOWN_FUNCTION,i,1);
				if (value != 0)
				{
					D[Di][Dj] = value;
					check_again = true;
if (DEBUG >= BASIC)
	debug << "S_switchfill:     set D[" << Di << "][" << Dj << "] = " << value << endl;					
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "S_switchfill:     no available value to complete row " << i << " in D" << endl;
					goto contradiction_bailout;
				}						
			}
		}
		
		/* check whether the current known values contradict S being a permuation */
		for (int k=1; k <= n; k++)
		{

if (DEBUG >= BASIC)
	debug << "S_switchfill:   permutation check for U[i][j]=" << k << endl;
	
			// for each U[i][j]=k, D[j][i] must be a permutation
			valarray<int> perm(0,n);
			for (int i=0; i< n; i++)
			for (int j=0; j< n; j++)
			{
				if (U[i][j] == k && D[j][i])
				{

if (DEBUG >= BASIC)
{
	debug << "S_switchfill:   U[" << i << "][" << j << "]=" << k << " and " 
	      << "D[" << j << "][" << i << "] = " << D[j][i] << endl;
}		  
	
					if (perm[D[j][i]-1])
					{
if (DEBUG >= BASIC)
	debug << "S_switchfill:   contradiction identified: D[j][i] is not a permutation for U[i][j]=" << k << endl;
						goto contradiction_bailout;
					}
					else
					{
if (DEBUG >= BASIC)
{
	debug << "S_switchfill:   setting perm[" << D[j][i]-1 << "] = 1" << endl;
}		  
						perm[D[j][i]-1] = 1;
					}
				}
			}
				
		}
		
	}	
	contradiction_bailout:	

if (DEBUG >= INTERMEDIATE)
{
	debug << "S_switchfill: returning (prototype) switch U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	return no_contradiction;
}

bool F_switchfill (Switch& S)
{
	matrix<int>& U = S.U;
	matrix<int>& D = S.D;
	
	bool no_contradiction = true;
	bool check_again = true;
	
	/* Check whether any zeros may be filled in from the up down and rule of four conditions */
	while (check_again)
	{
		check_again = false;
		
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill: starting check of triples a b c" << endl;
	debug << "F_switchfill: prototype switch: U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	
		for (int a=0; a < n; a++)
		for (int b=0; b < n; b++)
		for (int c=0; c < n; c++)
		{
		
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill: a = " << a << " b = " << b << " c = " << c << endl;
	
			/* the up interchange requires a^{bc_b} = a^{cb^c}; that is U[D[b][c]][U[b][a]] == U[U[c][b]][U[c][a]] */
			if (D[b][c] && U[b][a] && U[c][b] && U[c][a])
			{
				int lhs = U[D[b][c]-1][U[b][a]-1]; 
				int rhs = U[U[c][b]-1][U[c][a]-1]; 

if (DEBUG >= DETAIL)
{
	debug << "F_switchfill:   up interchange lhs: D[" << b << "][" << c << "]-1 = " << D[b][c]-1
	      << " U[" << b << "][" << a << "]-1 = " << U[b][a]-1 << endl;
	debug << "F_switchfill:   up interchange rhs: U[" << c << "][" << b << "]-1 = " << U[c][b]-1
          << " U[" << c << "][" << c << "]-1 = " << U[c][a]-1 << endl;
}

if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:   up interchange check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(U,U[c][b]-1,lhs))
					{
						U[U[c][b]-1][U[c][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting U[" << U[c][b]-1 << "][" << U[c][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << U[c][b]-1 << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(U,D[b][c]-1,rhs))
					{
						U[D[b][c]-1][U[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting U[" << D[b][c]-1 << "][" << U[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << c << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}								
			}
		
			/* the down interchange requires a_{bc^b} = a_{cb_c}; that is, D[U[b][c]][D[b][a]] == D[D[c][b]][D[c][a]] */
			if (U[b][c] && D[b][a] && D[c][b] && D[c][a])
			{
				int lhs = D[U[b][c]-1][D[b][a]-1];
				int rhs = D[D[c][b]-1][D[c][a]-1];
				
if (DEBUG >= DETAIL)
{
	debug << "F_switchfill:   down interchange lhs: U[" << b << "][" << c << "]-1 = " << U[b][c]-1
	      << " D[" << b << "][" << a << "]-1 = " << D[b][a]-1 << endl;
	debug << "F_switchfill:   down interchange rhs: D[" << c << "][" << b << "]-1 = " << D[c][b]-1
          << " D[" << c << "][" << a << "]-1 = " << D[c][a]-1 << endl;
}

if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:   down interchange check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(D,D[c][b]-1,lhs))
					{
						D[D[c][b]-1][D[c][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting D[" << D[c][b]-1 << "][" << D[c][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << D[c][b]-1 << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(D,U[b][c]-1,rhs))
					{
						D[U[b][c]-1][D[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting D[" << U[b][c]-1 << "][" << D[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << c << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}						
			}
			
			/* the rule of four requires that a_b^{c_b} = a^c_{b^c}; that is, U[D[b][c]][D[b][a]] == D[U[c][b]][U[c][a]] */
			if (D[b][c] && D[b][a] && U[c][b] && U[c][a])
			{
				int lhs = U[D[b][c]-1][D[b][a]-1];
				int rhs = D[U[c][b]-1][U[c][a]-1];
				
if (DEBUG >= DETAIL)
{
	debug << "F_switchfill:   rule of four lhs: D[" << b << "][" << c << "]-1 = " << D[b][c]-1
	      << " D[" << b << "][" << a << "]-1] = " << D[b][a]-1 << endl;
	debug << "F_switchfill:   rule of four rhs: U[" << c << "][" << c << "]-1 = " << U[c][b] 
	      << " U[" << c << "][" << a << "]-1 = " << U[c][a]-1 << endl;
}
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:   rule of five check lhs = " << lhs << " rhs = " << rhs << endl;
			
				if (lhs != 0 && rhs != 0)
				{
					if (lhs != rhs)
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     contradiction identified: lhs != rhs" << endl;						
						goto contradiction_bailout;
					}
				}
				else if (lhs != 0 && rhs == 0)
				{
					if (valid_row_perm_value(D,U[c][b]-1,lhs))
					{
						D[U[c][b]-1][U[c][a]-1] = lhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting D[" << U[c][b]-1 << "][" << U[c][a]-1 << "] = " << lhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << lhs << " already present in row " 
	      << U[D[a][c]-1][b]-1 << " of D" << endl;
}
						goto contradiction_bailout;
					}
				}
				else if (lhs == 0 && rhs != 0)
				{
					if (valid_row_perm_value(U,D[b][c]-1,rhs))
					{
						U[D[b][c]-1][D[b][a]-1] = rhs;
						check_again = true;
if (DEBUG >= INTERMEDIATE)
	debug << "F_switchfill:     setting U[" << D[b][c]-1 << "][" << D[b][a]-1 << "] = " << rhs << endl;
					}
					else
					{
						no_contradiction = false;
if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill:     contradiction identified: " << rhs << " already present in row " 
	      << D[U[a][b]-1][c]-1 << " of U" << endl;
}
						goto contradiction_bailout;
					}
				}						
			}
						
		}
		
		/* Check whether any rows have a single non-zero entry 
		   and fill in based on the permutation requirement */
		for (int i=0; i<n; i++)
		{
			int U_row_count = 0;
			int D_row_count = 0;
			int Ui, Uj, Di, Dj;
			
			for (int j=0; j < n && U_row_count < 2; j++)
			{
				if (U[i][j] == 0)
				{
					Ui=i;
					Uj=j;
					U_row_count++;
				}
			}
			
			if (U_row_count == 1)
			{
if (DEBUG >= BASIC)
	debug << "F_switchfill:   row " << i << " of U contains a single zero" << endl;					
				int value = avail(S,UP_FUNCTION,i,1);
				if (value != 0)
				{
					U[Ui][Uj] = value;
					check_again = true;
if (DEBUG >= BASIC)
	debug << "F_switchfill:   set U[" << Ui << "][" << Uj << "] = " << value << endl;					
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "F_switchfill:   no available value to complete row " << i << " in U" << endl;
					goto contradiction_bailout;
				}
					
			}
			
			for (int j=0; j < n && D_row_count < 2; j++)
			{
				if(D[i][j] == 0)
				{
					Di=i;
					Dj=j;
					D_row_count++;
				}
			}
			
			if (D_row_count == 1)
			{

if (DEBUG >= BASIC)
	debug << "F_switchfill:   row " << i << " of D contains a single zero" << endl;					

				int value = avail(S,DOWN_FUNCTION,i,1);
				if (value != 0)
				{
					D[Di][Dj] = value;
					check_again = true;
if (DEBUG >= BASIC)
	debug << "F_switchfill:     set D[" << Di << "][" << Dj << "] = " << value << endl;					
				}
				else
				{
if (DEBUG >= BASIC)
	debug << "F_switchfill:     no available value to complete row " << i << " in D" << endl;
					goto contradiction_bailout;
				}						
			}
		}
		
		/* check whether the current known values contradict S being a permuation */
		for (int k=1; k <= n; k++)
		{

if (DEBUG >= BASIC)
	debug << "F_switchfill:   permutation check for U[i][j]=" << k << endl;
	
			// for each U[i][j]=k, D[j][i] must be a permutation
			valarray<int> perm(0,n);
			for (int i=0; i< n; i++)
			for (int j=0; j< n; j++)
			{
				if (U[i][j] == k && D[j][i])
				{

if (DEBUG >= BASIC)
{
	debug << "F_switchfill:   U[" << i << "][" << j << "]=" << k << " and " 
	      << "D[" << j << "][" << i << "] = " << D[j][i] << endl;
}		  
	
					if (perm[D[j][i]-1])
					{
if (DEBUG >= BASIC)
	debug << "F_switchfill:   contradiction identified: D[j][i] is not a permutation for U[i][j]=" << k << endl;
						goto contradiction_bailout;
					}
					else
					{
if (DEBUG >= BASIC)
{
	debug << "F_switchfill:   setting perm[" << D[j][i]-1 << "] = 1" << endl;
}		  
						perm[D[j][i]-1] = 1;
					}
				}
			}
				
		}
		
	}	
	contradiction_bailout:	

if (DEBUG >= INTERMEDIATE)
{
	debug << "F_switchfill: returning (prototype) switch U = ";
	display (U, debug, TeX_OUTPUT_DISPLAY);
	debug << " D = ";
	display (D, debug, TeX_OUTPUT_DISPLAY);
	debug << endl;	
}
	return no_contradiction;
}

bool switchfill (Switch& S)
{
	if (S_YANG_BAXTER_SEARCH)
		return S_switchfill(S);
	else
		return F_switchfill(S);
}

void display(matrix<int>& M, ostream& os, bool TeX)
{
	if (TeX)
		os << "$$\\pmatrix{";
		
	for (int i=0; i< n; i++)
	{
		for (int j=0; j< n; j++)
			os << (NUMBER_FROM_ZERO? M[i][j]-1: M[i][j]) << ((TeX && j < n-1)? " & ": " ");
			
		if (i < n-1)
			os << (TeX? " \\cr ": " ");
	}

	if (TeX)
		os << "}$$";
}

/* valid_row_perm_value returns true if value does not appear in the 
   specified row of M and false otherwise.
*/
bool valid_row_perm_value(matrix<int>& M, int row, int value)
{
	for (int i=0; i< n; i++)
	{
		if (M[row][i] == value)
			return false;
	}
	return true;
}
